import os

import numpy as np

from yt.testing import (
    assert_almost_equal,
    assert_equal,
    assert_rel_equal,
    requires_file,
    requires_module,
)
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.answer_testing.framework import data_dir_load
from yt.utilities.cosmology import Cosmology
from yt.utilities.on_demand_imports import _yaml as yaml

local_dir = os.path.dirname(os.path.abspath(__file__))


def z_from_t_analytic(my_time, hubble_constant=0.7, omega_matter=0.3, omega_lambda=0.7):
    """
    Compute the redshift from time after the big bang.  This is based on
    Enzo's CosmologyComputeExpansionFactor.C, but altered to use physical
    units.
    """

    hubble_constant = YTQuantity(hubble_constant, "100*km/s/Mpc")
    omega_curvature = 1.0 - omega_matter - omega_lambda

    OMEGA_TOLERANCE = 1e-5
    ETA_TOLERANCE = 1.0e-10

    # Convert the time to Time * H0.

    if not isinstance(my_time, YTArray):
        my_time = YTArray(my_time, "s")

    t0 = (my_time.in_units("s") * hubble_constant.in_units("1/s")).to_ndarray()

    # For a flat universe with omega_matter = 1, it's easy.

    if np.fabs(omega_matter - 1) < OMEGA_TOLERANCE and omega_lambda < OMEGA_TOLERANCE:
        a = np.power(1.5 * t0, 2.0 / 3.0)

    # For omega_matter < 1 and omega_lambda == 0 see
    # Peebles 1993, eq. 13-3, 13-10.
    # Actually, this is a little tricky since we must solve an equation
    # of the form eta - np.sinh(eta) + x = 0..

    elif omega_matter < 1 and omega_lambda < OMEGA_TOLERANCE:
        x = 2 * t0 * np.power(1.0 - omega_matter, 1.5) / omega_matter

        # Compute eta in a three step process, first from a third-order
        # Taylor expansion of the formula above, then use that in a fifth-order
        # approximation.  Then finally, iterate on the formula itself, solving for
        # eta.  This works well because parts 1 & 2 are an excellent approximation
        # when x is small and part 3 converges quickly when x is large.

        eta = np.power(6 * x, 1.0 / 3.0)  # part 1
        eta = np.power(120 * x / (20 + eta * eta), 1.0 / 3.0)  # part 2
        mask = np.ones(eta.size, dtype=bool)
        max_iter = 1000
        for i in range(max_iter):  # part 3
            eta_old = eta[mask]
            eta[mask] = np.arcsinh(eta[mask] + x[mask])
            mask[mask] = np.fabs(eta[mask] - eta_old) >= ETA_TOLERANCE
            if not mask.any():
                break
            if i == max_iter - 1:
                raise RuntimeError("No convergence after %d iterations." % i)

        # Now use eta to compute the expansion factor (eq. 13-10, part 2).

        a = omega_matter / (2.0 * (1.0 - omega_matter)) * (np.cosh(eta) - 1.0)

    # For flat universe, with non-zero omega_lambda, see eq. 13-20.

    elif np.fabs(omega_curvature) < OMEGA_TOLERANCE and omega_lambda > OMEGA_TOLERANCE:
        a = np.power(omega_matter / (1 - omega_matter), 1.0 / 3.0) * np.power(
            np.sinh(1.5 * np.sqrt(1.0 - omega_matter) * t0), 2.0 / 3.0
        )

    else:
        raise NotImplementedError

    redshift = (1.0 / a) - 1.0

    return redshift


def t_from_z_analytic(z, hubble_constant=0.7, omega_matter=0.3, omega_lambda=0.7):
    """
    Compute the age of the Universe from redshift.  This is based on Enzo's
    CosmologyComputeTimeFromRedshift.C, but altered to use physical units.
    """

    hubble_constant = YTQuantity(hubble_constant, "100*km/s/Mpc")
    omega_curvature = 1.0 - omega_matter - omega_lambda

    # For a flat universe with omega_matter = 1, things are easy.

    if omega_matter == 1.0 and omega_lambda == 0.0:
        t0 = 2.0 / 3.0 / np.power(1 + z, 1.5)

    # For omega_matter < 1 and omega_lambda == 0 see
    # Peebles 1993, eq. 13-3, 13-10.

    elif omega_matter < 1 and omega_lambda == 0:
        eta = np.arccosh(1 + 2 * (1 - omega_matter) / omega_matter / (1 + z))
        t0 = (
            omega_matter
            / (2 * np.power(1.0 - omega_matter, 1.5))
            * (np.sinh(eta) - eta)
        )

    # For flat universe, with non-zero omega_lambda, see eq. 13-20.

    elif np.fabs(omega_curvature) < 1.0e-3 and omega_lambda != 0:
        t0 = (
            2.0
            / 3.0
            / np.sqrt(1 - omega_matter)
            * np.arcsinh(
                np.sqrt((1 - omega_matter) / omega_matter) / np.power(1 + z, 1.5)
            )
        )

    else:
        raise NotImplementedError(f"{hubble_constant}, {omega_matter}, {omega_lambda}")

    # Now convert from Time * H0 to time.

    my_time = t0 / hubble_constant

    return my_time


def test_z_t_roundtrip():
    """
    Make sure t_from_z and z_from_t are consistent.

    """

    co = Cosmology()
    # random sample in log(a) from -6 to 6
    my_random = np.random.RandomState(6132305)
    la = 12 * my_random.random_sample(10000) - 6
    z1 = 1 / np.power(10, la) - 1
    t = co.t_from_z(z1)
    z2 = co.z_from_t(t)
    assert_rel_equal(z1, z2, 4)


def test_z_t_analytic():
    """
    Test z/t conversions against analytic solutions.
    """

    cosmos = (
        {"hubble_constant": 0.7, "omega_matter": 0.3, "omega_lambda": 0.7},
        {"hubble_constant": 0.7, "omega_matter": 1.0, "omega_lambda": 0.0},
        {"hubble_constant": 0.7, "omega_matter": 0.3, "omega_lambda": 0.0},
    )

    for cosmo in cosmos:
        omega_curvature = 1 - cosmo["omega_matter"] - cosmo["omega_lambda"]
        co = Cosmology(omega_curvature=omega_curvature, **cosmo)
        # random sample in log(a) from -6 to 6
        my_random = np.random.RandomState(10132324)
        la = 12 * my_random.random_sample(1000) - 6
        z = 1 / np.power(10, la) - 1

        t_an = t_from_z_analytic(z, **cosmo).to("Gyr")
        t_co = co.t_from_z(z).to("Gyr")

        assert_rel_equal(
            t_an,
            t_co,
            4,
            err_msg=f"t_from_z does not match analytic version for cosmology {cosmo}.",
        )

        # random sample in log(t/t0) from -3 to 1
        t0 = np.power(10, 4 * my_random.random_sample(1000) - 3)
        t = (t0 / co.hubble_constant).to("Gyr")

        z_an = z_from_t_analytic(t, **cosmo)
        z_co = co.z_from_t(t)

        # compare scale factors since z approaches 0
        assert_rel_equal(
            1 / (1 + z_an),
            1 / (1 + z_co),
            5,
            err_msg=f"z_from_t does not match analytic version for cosmology {cosmo}.",
        )


def test_dark_factor():
    """
    Test that dark factor returns same value for when not
    being used and when w_0 = -1 and w_z = 0.
    """

    co = Cosmology(w_0=-1, w_a=0, use_dark_factor=False)

    assert_equal(co.get_dark_factor(0), 1.0)
    co.use_dark_factor = True
    assert_equal(co.get_dark_factor(0), 1.0)


@requires_module("yaml")
def test_cosmology_calculator_answers():
    """
    Test cosmology calculator functions against previously calculated values.
    """

    fn = os.path.join(local_dir, "cosmology_answers.yml")
    with open(fn) as fh:
        data = yaml.load(fh, Loader=yaml.FullLoader)

    cosmologies = data["cosmologies"]
    functions = data["functions"]

    for cname, copars in cosmologies.items():
        omega_curvature = (
            1
            - copars["omega_matter"]
            - copars["omega_lambda"]
            - copars["omega_radiation"]
        )

        cosmology = Cosmology(omega_curvature=omega_curvature, **copars)

        for fname, finfo in functions.items():
            func = getattr(cosmology, fname)
            args = finfo.get("args", [])
            val = func(*args)
            units = finfo.get("units")
            if units is not None:
                val.convert_to_units(units)
            val = float(val)

            err_msg = (
                "{} answer has changed for {} cosmology, old: {:f}, new: {:f}.".format(
                    fname,
                    cname,
                    finfo["answers"][cname],
                    val,
                )
            )
            assert_almost_equal(val, finfo["answers"][cname], 10, err_msg=err_msg)


enzotiny = "enzo_tiny_cosmology/DD0020/DD0020"


@requires_file(enzotiny)
def test_dataset_cosmology_calculator():
    """
    Test datasets's cosmology calculator against standalone.
    """

    ds = data_dir_load(enzotiny)

    co = Cosmology(
        hubble_constant=ds.hubble_constant,
        omega_matter=ds.omega_matter,
        omega_lambda=ds.omega_lambda,
    )

    v1 = ds.cosmology.comoving_radial_distance(1, 5).to("Mpccm").v
    v2 = co.comoving_radial_distance(1, 5).to("Mpccm").v
    assert_equal(v1, v2)



import os
import shutil
import tempfile
from unittest import TestCase

import numpy as np

from yt.testing import assert_fname, fake_random_ds, fake_vr_orientation_test_ds
from yt.visualization.volume_rendering.api import (
    create_scene,
    create_volume_source,
    volume_render,
)


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class RotationTest(TestCase):

    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def test_rotation(self):
        ds = fake_random_ds(32)
        ds2 = fake_random_ds(32)
        dd = ds.sphere(ds.domain_center, ds.domain_width[0] / 2)
        dd2 = ds2.sphere(ds2.domain_center, ds2.domain_width[0] / 2)

        im, sc = volume_render(dd, field=("gas", "density"))
        im.write_png("test.png")

        vol = sc.get_source(0)
        tf = vol.transfer_function
        tf.clear()
        mi, ma = dd.quantities.extrema(("gas", "density"))
        mi = np.log10(mi)
        ma = np.log10(ma)
        mi_bound = ((ma - mi) * (0.10)) + mi
        ma_bound = ((ma - mi) * (0.90)) + mi
        tf.map_to_colormap(mi_bound, ma_bound, scale=0.01, colormap="Blues_r")

        vol2 = create_volume_source(dd2, field=("gas", "density"))
        sc.add_source(vol2)

        tf = vol2.transfer_function
        tf.clear()
        mi, ma = dd2.quantities.extrema(("gas", "density"))
        mi = np.log10(mi)
        ma = np.log10(ma)
        mi_bound = ((ma - mi) * (0.10)) + mi
        ma_bound = ((ma - mi) * (0.90)) + mi
        tf.map_to_colormap(mi_bound, ma_bound, scale=0.01, colormap="Reds_r")
        fname = "test_scene.pdf"
        sc.save(fname, sigma_clip=6.0)
        assert_fname(fname)

        fname = "test_rot.png"
        sc.camera.pitch(np.pi)
        sc.render()
        sc.save(fname, sigma_clip=6.0, render=False)
        assert_fname(fname)


def test_annotations():
    from matplotlib.image import imread

    curdir = os.getcwd()
    tmpdir = tempfile.mkdtemp()
    os.chdir(tmpdir)
    ds = fake_vr_orientation_test_ds(N=16)
    sc = create_scene(ds)
    sc.annotate_axes()
    sc.annotate_domain(ds)
    sc.render()
    # ensure that there are actually red, green, blue, and white pixels
    # in the image. see Issue #1595
    im = sc._last_render
    for c in ([1, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1], [1, 1, 1, 1]):
        assert np.where((im == c).all(axis=-1))[0].shape[0] > 0
    sc[0].tfh.tf.add_layers(10, colormap="cubehelix")
    sc.save_annotated(
        "test_scene_annotated.png",
        text_annotate=[[(0.1, 1.05), "test_string"]],
    )
    image = imread("test_scene_annotated.png")
    assert image.shape == sc.camera.resolution + (4,)
    os.chdir(curdir)
    shutil.rmtree(tmpdir)



cdef class OrthoRaySelector(SelectorObject):

    cdef public np.uint8_t px_ax
    cdef public np.uint8_t py_ax
    cdef public np.float64_t px
    cdef public np.float64_t py
    cdef public int axis

    def __init__(self, dobj):
        self.axis = dobj.axis
        self.px_ax = dobj.px_ax
        self.py_ax = dobj.py_ax
        self.px = dobj.px
        self.py = dobj.py

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask_regular_grid(self, gobj):
        cdef np.ndarray[np.uint8_t, ndim=3] mask
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        cdef int i, j, k
        cdef int total = 0
        cdef int this_level = 0
        cdef int ind[3][2]
        cdef np.int32_t level = gobj.Level
        _ensure_code(gobj.LeftEdge)
        _ensure_code(gobj.RightEdge)
        _ensure_code(gobj.dds)

        if level < self.min_level or level > self.max_level:
            return None
        else:
            child_mask = gobj.child_mask
            mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
            if level == self.max_level:
                this_level = 1
            ind[self.axis][0] = 0
            ind[self.axis][1] = gobj.ActiveDimensions[self.axis]
            ind[self.px_ax][0] = \
                <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /
                       gobj.dds[self.px_ax])
            ind[self.px_ax][1] = ind[self.px_ax][0] + 1
            ind[self.py_ax][0] = \
                <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /
                       gobj.dds[self.py_ax])
            ind[self.py_ax][1] = ind[self.py_ax][0] + 1

            with nogil:
                for i in range(ind[0][0], ind[0][1]):
                    for j in range(ind[1][0], ind[1][1]):
                        for k in range(ind[2][0], ind[2][1]):
                            if this_level == 1 or child_mask[i, j, k]:
                                mask[i, j, k] = 1
                                total += 1
            if total == 0: return None
            return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
        if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \
           self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
           self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
           self.py <  pos[self.py_ax] + 0.5*dds[self.py_ax]:
            return 1
        return 0

    cdef int select_point(self, np.float64_t pos[3]) nogil:
        # two 0-volume constructs don't intersect
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
        cdef np.float64_t dx = self.periodic_difference(
            pos[self.px_ax], self.px, self.px_ax)
        cdef np.float64_t dy = self.periodic_difference(
            pos[self.py_ax], self.py, self.py_ax)
        if dx*dx + dy*dy < radius*radius:
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \
           left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \
           left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
            return 2 # a box of non-zero volume can't be inside a ray
        return 0

    def _hash_vals(self):
        return (("px_ax", self.px_ax),
                ("py_ax", self.py_ax),
                ("px", self.px),
                ("py", self.py),
                ("axis", self.axis))

    def _get_state_attnames(self):
        return ("px_ax", "py_ax", "px", "py", "axis")

ortho_ray_selector = OrthoRaySelector



import numpy as np

from yt.frontends.enzo.api import EnzoDataset
from yt.frontends.enzo.fields import NODAL_FLAGS
from yt.testing import (
    assert_allclose_units,
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
    requires_file,
    units_override_check,
)
from yt.utilities.answer_testing.framework import (
    big_patch_amr,
    data_dir_load,
    requires_ds,
    small_patch_amr,
)
from yt.visualization.plot_window import SlicePlot

_fields = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
    ("gas", "velocity_divergence"),
)

two_sphere_test = "ActiveParticleTwoSphere/DD0011/DD0011"
active_particle_cosmology = "ActiveParticleCosmology/DD0046/DD0046"
ecp = "enzo_cosmology_plus/DD0046/DD0046"
g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"
enzotiny = "enzo_tiny_cosmology/DD0046/DD0046"
toro1d = "ToroShockTube/DD0001/data0001"
kh2d = "EnzoKelvinHelmholtz/DD0011/DD0011"
mhdctot = "MHDCTOrszagTang/DD0004/data0004"
dnz = "DeeplyNestedZoom/DD0025/data0025"
p3mini = "PopIII_mini/DD0034/DD0034"


def color_conservation(ds):
    species_names = ds.field_info.species_names
    dd = ds.all_data()
    dens_yt = dd[("gas", "density")].copy()
    # Enumerate our species here
    for s in sorted(species_names):
        if s == "El":
            continue
        dens_yt -= dd[f"{s}_density"]
    dens_yt -= dd[("enzo", "Metal_Density")]
    delta_yt = np.abs(dens_yt / dd[("gas", "density")])
    # Now we compare color conservation to Enzo's color conservation
    dd = ds.all_data()
    dens_enzo = dd[("enzo", "Density")].copy()
    for f in sorted(ds.field_list):
        ff = f[1]
        if not ff.endswith("_Density"):
            continue
        start_strings = [
            "Electron_",
            "SFR_",
            "Forming_Stellar_",
            "Dark_Matter",
            "Star_Particle_",
        ]
        if any([ff.startswith(ss) for ss in start_strings]):
            continue
        dens_enzo -= dd[f]
    delta_enzo = np.abs(dens_enzo / dd[("enzo", "Density")])
    np.testing.assert_almost_equal(delta_yt, delta_enzo)


def check_color_conservation(ds):
    species_names = ds.field_info.species_names
    dd = ds.all_data()
    dens_yt = dd[("gas", "density")].copy()
    # Enumerate our species here
    for s in sorted(species_names):
        if s == "El":
            continue
        dens_yt -= dd[f"{s}_density"]
    dens_yt -= dd[("enzo", "Metal_Density")]
    delta_yt = np.abs(dens_yt / dd[("gas", "density")])

    # Now we compare color conservation to Enzo's color conservation
    dd = ds.all_data()
    dens_enzo = dd[("enzo", "Density")].copy()
    for f in sorted(ds.field_list):
        ff = f[1]
        if not ff.endswith("_Density"):
            continue
        start_strings = [
            "Electron_",
            "SFR_",
            "Forming_Stellar_",
            "Dark_Matter",
            "Star_Particle_",
        ]
        if any([ff.startswith(ss) for ss in start_strings]):
            continue
        dens_enzo -= dd[f]
    delta_enzo = np.abs(dens_enzo / dd[("enzo", "Density")])
    return assert_almost_equal, delta_yt, delta_enzo


m7 = "DD0010/moving7_0010"


@requires_ds(m7)
def test_moving7():
    ds = data_dir_load(m7)
    assert_equal(str(ds), "moving7_0010")
    for test in small_patch_amr(m7, _fields):
        test_moving7.__name__ = test.description
        yield test


@requires_ds(g30, big_data=True)
def test_galaxy0030():
    ds = data_dir_load(g30)
    yield check_color_conservation(ds)
    assert_equal(str(ds), "galaxy0030")
    for test in big_patch_amr(ds, _fields):
        test_galaxy0030.__name__ = test.description
        yield test
    assert_equal(ds.particle_type_counts, {"io": 1124453})


@requires_ds(toro1d)
def test_toro1d():
    ds = data_dir_load(toro1d)
    assert_equal(str(ds), "data0001")
    for test in small_patch_amr(ds, ds.field_list):
        test_toro1d.__name__ = test.description
        yield test


@requires_ds(kh2d)
def test_kh2d():
    ds = data_dir_load(kh2d)
    assert_equal(str(ds), "DD0011")
    for test in small_patch_amr(ds, ds.field_list):
        test_kh2d.__name__ = test.description
        yield test


@requires_ds(ecp, big_data=True)
def test_ecp():
    ds = data_dir_load(ecp)
    # Now we test our species fields
    yield check_color_conservation(ds)


@requires_file(enzotiny)
def test_units_override():
    units_override_check(enzotiny)


@requires_ds(ecp, big_data=True)
def test_nuclei_density_fields():
    ds = data_dir_load(ecp)
    ad = ds.all_data()
    assert_array_equal(
        ad[("gas", "H_nuclei_density")],
        (ad[("gas", "H_p0_number_density")] + ad[("gas", "H_p1_number_density")]),
    )
    assert_array_equal(
        ad[("gas", "He_nuclei_density")],
        (
            ad[("gas", "He_p0_number_density")]
            + ad[("gas", "He_p1_number_density")]
            + ad[("gas", "He_p2_number_density")]
        ),
    )


@requires_file(enzotiny)
def test_EnzoDataset():
    assert isinstance(data_dir_load(enzotiny), EnzoDataset)


@requires_file(two_sphere_test)
@requires_file(active_particle_cosmology)
def test_active_particle_datasets():
    two_sph = data_dir_load(two_sphere_test)
    assert "AccretingParticle" in two_sph.particle_types_raw
    assert "io" not in two_sph.particle_types_raw
    assert "all" in two_sph.particle_types
    assert "nbody" in two_sph.particle_types
    assert_equal(len(two_sph.particle_unions), 2)
    pfields = [
        "GridID",
        "creation_time",
        "dynamical_time",
        "identifier",
        "level",
        "metallicity",
        "particle_mass",
    ]
    pfields += [f"particle_position_{d}" for d in "xyz"]
    pfields += [f"particle_velocity_{d}" for d in "xyz"]

    acc_part_fields = [("AccretingParticle", pf) for pf in ["AccretionRate"] + pfields]

    real_acc_part_fields = sorted(
        f for f in two_sph.field_list if f[0] == "AccretingParticle"
    )
    assert_equal(acc_part_fields, real_acc_part_fields)

    apcos = data_dir_load(active_particle_cosmology)
    assert_equal(["CenOstriker", "DarkMatter"], apcos.particle_types_raw)
    assert "all" in apcos.particle_unions
    assert "nbody" in apcos.particle_unions

    apcos_fields = [("CenOstriker", pf) for pf in pfields]

    real_apcos_fields = sorted(f for f in apcos.field_list if f[0] == "CenOstriker")

    assert_equal(apcos_fields, real_apcos_fields)

    assert_equal(
        apcos.particle_type_counts, {"CenOstriker": 899755, "DarkMatter": 32768}
    )


@requires_file(mhdctot)
def test_face_centered_mhdct_fields():
    ds = data_dir_load(mhdctot)

    ad = ds.all_data()
    grid = ds.index.grids[0]

    for field, flag in NODAL_FLAGS.items():
        dims = ds.domain_dimensions
        assert_equal(ad[field].shape, (dims.prod(), 2 * sum(flag)))
        assert_equal(grid[field].shape, tuple(dims) + (2 * sum(flag),))

    # Average of face-centered fields should be the same as cell-centered field
    assert (ad[("enzo", "BxF")].sum(axis=-1) / 2 == ad[("enzo", "Bx")]).all()
    assert (ad[("enzo", "ByF")].sum(axis=-1) / 2 == ad[("enzo", "By")]).all()
    assert (ad[("enzo", "BzF")].sum(axis=-1) / 2 == ad[("enzo", "Bz")]).all()


@requires_file(dnz)
def test_deeply_nested_zoom():
    ds = data_dir_load(dnz)

    # carefully chosen to just barely miss a grid in the middle of the image
    center = [0.4915073260199302, 0.5052605316800006, 0.4905805557500548]

    plot = SlicePlot(ds, "z", "density", width=(0.001, "pc"), center=center)

    image = plot.frb[("gas", "density")]

    assert (image > 0).all()

    v, c = ds.find_max(("gas", "density"))

    assert_allclose_units(v, ds.quan(0.005878286377124154, "g/cm**3"))

    c_actual = [0.49150732540021, 0.505260532936791, 0.49058055816398]
    c_actual = ds.arr(c_actual, "code_length")
    assert_allclose_units(c, c_actual)

    assert_equal(max(g[("gas", "density")].max() for g in ds.index.grids), v)


@requires_file(kh2d)
def test_2d_grid_shape():
    # see issue #1601
    # we want to make sure that accessing data on a grid object
    # returns a 3D array with a dummy dimension.
    ds = data_dir_load(kh2d)
    g = ds.index.grids[1]
    assert g[("gas", "density")].shape == (128, 100, 1)


@requires_file(p3mini)
def test_nonzero_omega_radiation():
    """
    Test support for non-zero omega_radiation cosmologies.
    """
    ds = data_dir_load(p3mini)

    assert_equal(ds.omega_radiation, ds.cosmology.omega_radiation)

    tratio = ds.current_time / ds.cosmology.t_from_z(ds.current_redshift)
    assert_almost_equal(
        tratio,
        1,
        4,
        err_msg="Simulation time not consistent with cosmology calculator.",
    )



import numpy as np

from yt.utilities.lib.misc_utilities import (
    obtain_position_vector,
    obtain_relative_velocity_vector,
)

from .derived_field import ValidateParameter
from .field_plugin_registry import register_field_plugin
from .vector_operations import create_magnitude_field


@register_field_plugin
def setup_angular_momentum(registry, ftype="gas", slice_info=None):
    # Angular momentum defined here needs to be consistent with
    # _particle_specific_angular_momentum in particle_fields.py
    unit_system = registry.ds.unit_system

    def _specific_angular_momentum_x(field, data):
        xv, yv, zv = obtain_relative_velocity_vector(data)
        rv = obtain_position_vector(data)
        units = rv.units
        rv = np.rollaxis(rv, 0, len(rv.shape))
        rv = data.ds.arr(rv, units=units)
        return rv[..., 1] * zv - rv[..., 2] * yv

    def _specific_angular_momentum_y(field, data):
        xv, yv, zv = obtain_relative_velocity_vector(data)
        rv = obtain_position_vector(data)
        units = rv.units
        rv = np.rollaxis(rv, 0, len(rv.shape))
        rv = data.ds.arr(rv, units=units)
        return rv[..., 2] * xv - rv[..., 0] * zv

    def _specific_angular_momentum_z(field, data):
        xv, yv, zv = obtain_relative_velocity_vector(data)
        rv = obtain_position_vector(data)
        units = rv.units
        rv = np.rollaxis(rv, 0, len(rv.shape))
        rv = data.ds.arr(rv, units=units)
        return rv[..., 0] * yv - rv[..., 1] * xv

    registry.add_field(
        (ftype, "specific_angular_momentum_x"),
        sampling_type="local",
        function=_specific_angular_momentum_x,
        units=unit_system["specific_angular_momentum"],
        validators=[ValidateParameter("center"), ValidateParameter("bulk_velocity")],
    )

    registry.add_field(
        (ftype, "specific_angular_momentum_y"),
        sampling_type="local",
        function=_specific_angular_momentum_y,
        units=unit_system["specific_angular_momentum"],
        validators=[ValidateParameter("center"), ValidateParameter("bulk_velocity")],
    )

    registry.add_field(
        (ftype, "specific_angular_momentum_z"),
        sampling_type="local",
        function=_specific_angular_momentum_z,
        units=unit_system["specific_angular_momentum"],
        validators=[ValidateParameter("center"), ValidateParameter("bulk_velocity")],
    )

    create_magnitude_field(
        registry,
        "specific_angular_momentum",
        unit_system["specific_angular_momentum"],
        ftype=ftype,
    )

    def _angular_momentum_x(field, data):
        return data[ftype, "mass"] * data[ftype, "specific_angular_momentum_x"]

    registry.add_field(
        (ftype, "angular_momentum_x"),
        sampling_type="local",
        function=_angular_momentum_x,
        units=unit_system["angular_momentum"],
        validators=[ValidateParameter("center"), ValidateParameter("bulk_velocity")],
    )

    def _angular_momentum_y(field, data):
        return data[ftype, "mass"] * data[ftype, "specific_angular_momentum_y"]

    registry.add_field(
        (ftype, "angular_momentum_y"),
        sampling_type="local",
        function=_angular_momentum_y,
        units=unit_system["angular_momentum"],
        validators=[ValidateParameter("center"), ValidateParameter("bulk_velocity")],
    )

    def _angular_momentum_z(field, data):
        return data[ftype, "mass"] * data[ftype, "specific_angular_momentum_z"]

    registry.add_field(
        (ftype, "angular_momentum_z"),
        sampling_type="local",
        function=_angular_momentum_z,
        units=unit_system["angular_momentum"],
        validators=[ValidateParameter("center"), ValidateParameter("bulk_velocity")],
    )

    create_magnitude_field(
        registry, "angular_momentum", unit_system["angular_momentum"], ftype=ftype
    )



import tempfile
from pathlib import Path

from yt.data_objects.static_output import Dataset
from yt.data_objects.time_series import DatasetSeries
from yt.testing import assert_raises
from yt.utilities.exceptions import YTUnidentifiedDataType
from yt.utilities.object_registries import output_type_registry


def test_pattern_expansion():
    file_list = [f"fake_data_file_{str(i).zfill(4)}" for i in range(10)]

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        for file in file_list:
            (tmp_path / file).touch()

        pattern = tmp_path / "fake_data_file_*"
        expected = [str(tmp_path / file) for file in file_list]
        found = DatasetSeries._get_filenames_from_glob_pattern(pattern)
        assert found == expected

        found2 = DatasetSeries._get_filenames_from_glob_pattern(Path(pattern))
        assert found2 == expected


def test_no_match_pattern():
    with tempfile.TemporaryDirectory() as tmpdir:
        pattern = Path(tmpdir).joinpath("fake_data_file_*")
        assert_raises(
            FileNotFoundError, DatasetSeries._get_filenames_from_glob_pattern, pattern
        )


def test_init_fake_dataseries():

    file_list = [f"fake_data_file_{str(i).zfill(4)}" for i in range(10)]
    with tempfile.TemporaryDirectory() as tmpdir:
        pfile_list = [Path(tmpdir) / file for file in file_list]
        sfile_list = [str(file) for file in pfile_list]
        for file in pfile_list:
            file.touch()
        pattern = Path(tmpdir) / "fake_data_file_*"

        # init from str pattern
        ts = DatasetSeries(pattern)
        assert ts._pre_outputs == sfile_list

        # init from Path pattern
        ppattern = Path(pattern)
        ts = DatasetSeries(ppattern)
        assert ts._pre_outputs == sfile_list

        # init form str list
        ts = DatasetSeries(sfile_list)
        assert ts._pre_outputs == sfile_list

        # init form Path list
        ts = DatasetSeries(pfile_list)
        assert ts._pre_outputs == pfile_list

        # rejected input type (str repr of a list) "[file1, file2, ...]"
        assert_raises(FileNotFoundError, DatasetSeries, str(file_list))

        # finally, check that ts[0] fails to actually load
        assert_raises(YTUnidentifiedDataType, ts.__getitem__, 0)

        class FakeDataset(Dataset):
            """A minimal loadable fake dataset subclass"""

            @classmethod
            def _is_valid(cls, *args, **kwargs):
                return True

            def _parse_parameter_file(self):
                return

            def _set_code_unit_attributes(self):
                return

            def set_code_units(self):
                self.current_time = 0
                return

            def _hash(self):
                return

            def _setup_classes(self):
                return

        try:
            ds = DatasetSeries(pattern)[0]
            assert isinstance(ds, FakeDataset)

            ts = DatasetSeries(pattern, my_unsupported_kwarg=None)

            assert_raises(TypeError, ts.__getitem__, 0)
            # the exact error message is supposed to be this
            # """__init__() got an unexpected keyword argument 'my_unsupported_kwarg'"""
            # but it's hard to check for within the framework
        finally:
            # tear down to avoid possible breakage in following tests
            output_type_registry.pop("FakeDataset")



from yt.frontends.art.api import ARTDataset
from yt.testing import (
    ParticleSelectionComparison,
    assert_almost_equal,
    assert_equal,
    requires_file,
    units_override_check,
)
from yt.units.yt_array import YTQuantity
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    PixelizedProjectionValuesTest,
    data_dir_load,
    requires_ds,
)

_fields = (
    ("gas", "density"),
    ("gas", "temperature"),
    ("all", "particle_mass"),
    ("all", "particle_position_x"),
    ("all", "particle_velocity_y"),
)

d9p = "D9p_500/10MpcBox_HartGal_csf_a0.500.d"
dmonly = "DMonly/PMcrs0.0100.DAT"


@requires_ds(d9p, big_data=True)
def test_d9p():
    ds = data_dir_load(d9p)
    ds.index
    assert_equal(str(ds), "10MpcBox_HartGal_csf_a0.500.d")
    dso = [None, ("sphere", ("max", (0.1, "unitary")))]
    for field in _fields:
        for axis in [0, 1]:
            for dobj_name in dso:
                for weight_field in [None, ("gas", "density")]:
                    if field[0] not in ds.particle_types:
                        yield PixelizedProjectionValuesTest(
                            d9p, axis, field, weight_field, dobj_name
                        )
                yield FieldValuesTest(
                    d9p,
                    field,
                    obj_type=dobj_name,
                    particle_type=field[0] in ds.particle_types,
                )


@requires_ds(d9p, big_data=True)
def test_d9p_global_values():
    ds = data_dir_load(d9p)
    ad = ds.all_data()
    # 'Ana' variable values output from the ART Fortran 'ANA' analysis code
    AnaNStars = 6255
    assert_equal(ad[("stars", "particle_type")].size, AnaNStars)
    assert_equal(ad[("specie4", "particle_type")].size, AnaNStars)

    # The *real* answer is 2833405, but yt misses one particle since it lives
    # on a domain boundary. See issue 814. When that is fixed, this test
    # will need to be updated
    AnaNDM = 2833404
    assert_equal(ad[("darkmatter", "particle_type")].size, AnaNDM)
    assert_equal(
        (
            ad[("specie0", "particle_type")].size
            + ad[("specie1", "particle_type")].size
            + ad[("specie2", "particle_type")].size
            + ad[("specie3", "particle_type")].size
        ),
        AnaNDM,
    )

    for spnum in range(5):
        npart_read = ad[f"specie{spnum}", "particle_type"].size
        npart_header = ds.particle_type_counts[f"specie{spnum}"]
        if spnum == 3:
            # see issue 814
            npart_read += 1
        assert_equal(npart_read, npart_header)

    AnaBoxSize = YTQuantity(7.1442196564, "Mpc")
    AnaVolume = YTQuantity(364.640074656, "Mpc**3")
    Volume = 1
    for i in ds.domain_width.in_units("Mpc"):
        assert_almost_equal(i, AnaBoxSize)
        Volume *= i
    assert_almost_equal(Volume, AnaVolume)

    AnaNCells = 4087490
    assert_equal(len(ad[("index", "cell_volume")]), AnaNCells)

    AnaTotDMMass = YTQuantity(1.01191786808255e14, "Msun")
    assert_almost_equal(
        ad[("darkmatter", "particle_mass")].sum().in_units("Msun"), AnaTotDMMass
    )

    AnaTotStarMass = YTQuantity(1776701.3990607238, "Msun")
    assert_almost_equal(
        ad[("stars", "particle_mass")].sum().in_units("Msun"), AnaTotStarMass
    )

    AnaTotStarMassInitial = YTQuantity(2423468.2801332865, "Msun")
    assert_almost_equal(
        ad[("stars", "particle_mass_initial")].sum().in_units("Msun"),
        AnaTotStarMassInitial,
    )

    AnaTotGasMass = YTQuantity(1.7826982029216785e13, "Msun")
    assert_almost_equal(ad[("gas", "cell_mass")].sum().in_units("Msun"), AnaTotGasMass)

    AnaTotTemp = YTQuantity(150219844793.3907, "K")  # just leaves
    assert_almost_equal(ad[("gas", "temperature")].sum().in_units("K"), AnaTotTemp)


@requires_file(d9p)
def test_ARTDataset():
    assert isinstance(data_dir_load(d9p), ARTDataset)


@requires_file(d9p)
def test_units_override():
    units_override_check(d9p)


@requires_file(dmonly)
def test_particle_selection():
    ds = data_dir_load(dmonly)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()



from yt.testing import assert_array_equal, assert_equal, requires_file
from yt.utilities.answer_testing.framework import (
    GenericArrayTest,
    data_dir_load,
    requires_ds,
)

out = "ExodusII/out.e"


@requires_file(out)
def test_out():
    ds = data_dir_load(out)
    field_list = [
        ("all", "conv_indicator"),
        ("all", "conv_marker"),
        ("all", "convected"),
        ("all", "diffused"),
        ("connect1", "conv_indicator"),
        ("connect1", "conv_marker"),
        ("connect1", "convected"),
        ("connect1", "diffused"),
        ("connect2", "conv_indicator"),
        ("connect2", "conv_marker"),
        ("connect2", "convected"),
        ("connect2", "diffused"),
    ]
    assert_equal(str(ds), "out.e")
    assert_equal(ds.dimensionality, 3)
    assert_equal(ds.current_time, 0.0)
    assert_array_equal(ds.parameters["nod_names"], ["convected", "diffused"])
    assert_equal(ds.parameters["num_meshes"], 2)
    assert_array_equal(ds.field_list, field_list)


out_s002 = "ExodusII/out.e-s002"


@requires_file(out_s002)
def test_out002():
    ds = data_dir_load(out_s002)
    field_list = [
        ("all", "conv_indicator"),
        ("all", "conv_marker"),
        ("all", "convected"),
        ("all", "diffused"),
        ("connect1", "conv_indicator"),
        ("connect1", "conv_marker"),
        ("connect1", "convected"),
        ("connect1", "diffused"),
        ("connect2", "conv_indicator"),
        ("connect2", "conv_marker"),
        ("connect2", "convected"),
        ("connect2", "diffused"),
    ]
    assert_equal(str(ds), "out.e-s002")
    assert_equal(ds.dimensionality, 3)
    assert_equal(ds.current_time, 2.0)
    assert_array_equal(ds.field_list, field_list)


gold = "ExodusII/gold.e"


@requires_file(gold)
def test_gold():
    ds = data_dir_load(gold)
    field_list = [("all", "forced"), ("connect1", "forced")]
    assert_equal(str(ds), "gold.e")
    assert_array_equal(ds.field_list, field_list)


big_data = "MOOSE_sample_data/mps_out.e"


@requires_ds(big_data)
def test_displacement_fields():
    displacement_dicts = [
        {"connect2": (5.0, [0.0, 0.0, 0.0])},
        {"connect1": (1.0, [1.0, 2.0, 3.0]), "connect2": (0.0, [0.0, 0.0, 0.0])},
    ]
    for disp in displacement_dicts:
        ds = data_dir_load(big_data, displacements=disp)
        for mesh in ds.index.meshes:

            def array_func():
                return mesh.connectivity_coords  # noqa: B023

            yield GenericArrayTest(ds, array_func, 12)



from math import pi

from yt.units.yt_array import YTQuantity
from yt.utilities.physical_ratios import (
    amu_grams,
    boltzmann_constant_erg_per_K,
    mass_earth_grams,
    mass_electron_grams,
    mass_hydrogen_grams,
    mass_jupiter_grams,
    mass_mars_grams,
    mass_mercury_grams,
    mass_neptune_grams,
    mass_saturn_grams,
    mass_sun_grams,
    mass_uranus_grams,
    mass_venus_grams,
    newton_cgs,
    planck_cgs,
    planck_charge_esu,
    planck_energy_erg,
    planck_length_cm,
    planck_mass_grams,
    planck_temperature_K,
    planck_time_s,
    speed_of_light_cm_per_s,
    standard_gravity_cm_per_s2,
)

mass_electron_cgs = YTQuantity(mass_electron_grams, "g")
mass_electron = mass_electron_cgs
me = mass_electron_cgs

amu_cgs = YTQuantity(amu_grams, "g")
amu = amu_cgs
Na = 1 / amu_cgs

mass_hydrogen_cgs = YTQuantity(mass_hydrogen_grams, "g")
mass_hydrogen = mass_hydrogen_cgs
mp = mass_hydrogen_cgs
mh = mp

# Velocities
speed_of_light_cgs = YTQuantity(speed_of_light_cm_per_s, "cm/s")
speed_of_light = speed_of_light_cgs
clight = speed_of_light_cgs
c = speed_of_light_cgs

# Cross Sections
# 8*pi/3 (alpha*hbar*c/(2*pi))**2
cross_section_thompson_cgs = YTQuantity(6.65245854533e-25, "cm**2")
cross_section_thompson = cross_section_thompson_cgs
thompson_cross_section = cross_section_thompson_cgs
sigma_thompson = cross_section_thompson_cgs

# Charge
charge_proton_cgs = YTQuantity(4.8032056e-10, "esu")
charge_proton = charge_proton_cgs
proton_charge = charge_proton_cgs
elementary_charge = charge_proton_cgs
qp = charge_proton_cgs

# Physical Constants
boltzmann_constant_cgs = YTQuantity(boltzmann_constant_erg_per_K, "erg/K")
boltzmann_constant = boltzmann_constant_cgs
kboltz = boltzmann_constant_cgs
kb = kboltz

gravitational_constant_cgs = YTQuantity(newton_cgs, "cm**3/g/s**2")
gravitational_constant = gravitational_constant_cgs
G = gravitational_constant_cgs

planck_constant_cgs = YTQuantity(planck_cgs, "erg*s")
planck_constant = planck_constant_cgs
hcgs = planck_constant_cgs
hbar = 0.5 * hcgs / pi

stefan_boltzmann_constant_cgs = YTQuantity(5.670373e-5, "erg/cm**2/s**1/K**4")
stefan_boltzmann_constant = stefan_boltzmann_constant_cgs

Tcmb = YTQuantity(2.726, "K")  # Current CMB temperature
CMB_temperature = Tcmb

# Solar System
mass_sun_cgs = YTQuantity(mass_sun_grams, "g")
mass_sun = mass_sun_cgs
solar_mass = mass_sun_cgs
msun = mass_sun_cgs
# Standish, E.M. (1995) "Report of the IAU WGAS Sub-Group on Numerical Standards",
# in Highlights of Astronomy (I. Appenzeller, ed.), Table 1,
# Kluwer Academic Publishers, Dordrecht.
# REMARK: following masses include whole systems (planet + moons)
mass_jupiter_cgs = YTQuantity(mass_jupiter_grams, "g")
mass_jupiter = mass_jupiter_cgs
jupiter_mas = mass_jupiter_cgs

mass_mercury_cgs = YTQuantity(mass_mercury_grams, "g")
mass_mercury = mass_mercury_cgs
mercury_mass = mass_mercury_cgs

mass_venus_cgs = YTQuantity(mass_venus_grams, "g")
mass_venus = mass_venus_cgs
venus_mass = mass_venus_cgs

mass_earth_cgs = YTQuantity(mass_earth_grams, "g")
mass_earth = mass_earth_cgs
earth_mass = mass_earth_cgs
mearth = mass_earth_cgs

mass_mars_cgs = YTQuantity(mass_mars_grams, "g")
mass_mars = mass_mars_cgs
mars_mass = mass_mars_cgs

mass_saturn_cgs = YTQuantity(mass_saturn_grams, "g")
mass_saturn = mass_saturn_cgs
saturn_mass = mass_saturn_cgs

mass_uranus_cgs = YTQuantity(mass_uranus_grams, "g")
mass_uranus = mass_uranus_cgs
uranus_mass = mass_uranus_cgs

mass_neptune_cgs = YTQuantity(mass_neptune_grams, "g")
mass_neptune = mass_neptune_cgs
neptune_mass = mass_neptune_cgs

# Planck units
m_pl = planck_mass = YTQuantity(planck_mass_grams, "g")
l_pl = planck_length = YTQuantity(planck_length_cm, "cm")
t_pl = planck_time = YTQuantity(planck_time_s, "s")
E_pl = planck_energy = YTQuantity(planck_energy_erg, "erg")
q_pl = planck_charge = YTQuantity(planck_charge_esu, "esu")
T_pl = planck_temperature = YTQuantity(planck_temperature_K, "K")

# MKS E&M units
mu_0 = YTQuantity(4.0e-7 * pi, "N/A**2")
eps_0 = (1.0 / (clight**2 * mu_0)).in_units("C**2/N/m**2")

# Misc
standard_gravity_cgs = YTQuantity(standard_gravity_cm_per_s2, "cm/s**2")
standard_gravity = standard_gravity_cgs



# These are some of the data object registries that are used in different places in the
# code. Not all of the self-registering objects are included in these.

# type hints are simplified as raw Type (instead of, e.g., Type[Dataset])
# to workaround circular imports
from typing import Dict, Type

# subclasses of yt.data_objects.analyzer_objects.AnalysisTask
analysis_task_registry: Dict[str, Type] = {}

# subclasses of yt.data_objects.data_containers.YTDataContainer
data_object_registry: Dict[str, Type] = {}

# suclasses of yt.data_objects.derived_quantity.DerivedQuantity
derived_quantity_registry: Dict[str, Type] = {}

# suclasses of yt.data_objects.static_outputs.Dataset
output_type_registry: Dict[str, Type] = {}

# subclasses of yt.data_objects.time_series.DatasetSeries
simulation_time_series_registry: Dict[str, Type] = {}



import yaml
from sympy import Matrix, MatrixSymbol, ccode, diff, symarray

# define some templates used below
fun_signature = """cdef void %s(double* fx,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil"""

fun_dec_template = fun_signature + " \n"
fun_def_template = (
    """@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) \n"""
    + fun_signature
    + ": \n"
)

jac_signature_3D = """cdef void %s(double* rcol,
                       double* scol,
                       double* tcol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil"""

jac_dec_template_3D = jac_signature_3D + " \n"
jac_def_template_3D = (
    """@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) \n"""
    + jac_signature_3D
    + ": \n"
)

jac_signature_2D = """cdef void %s(double* rcol,
                       double* scol,
                       double* x,
                       double* vertices,
                       double* phys_x) nogil"""
jac_dec_template_2D = jac_signature_2D + " \n"
jac_def_template_2D = (
    """@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) \n"""
    + jac_signature_2D
    + ": \n"
)

file_header = (
    "# This file contains auto-generated functions for sampling \n"
    + "# inside finite element solutions for various mesh types. \n"
    + "# To see how the code generation works in detail, see \n"
    + "# yt/utilities/mesh_code_generation.py. \n"
)


class MeshCodeGenerator:
    """

    A class for automatically generating the functions and jacobians used for
    sampling inside finite element calculations.

    """

    def __init__(self, mesh_data):
        """

        Mesh data should be a dictionary containing information about the type
        of elements used. See yt/utilities/mesh_types.yaml for more information.

        """
        self.mesh_type = mesh_data["mesh_type"]
        self.num_dim = mesh_data["num_dim"]
        self.num_vertices = mesh_data["num_vertices"]
        self.num_mapped_coords = mesh_data["num_mapped_coords"]

        x = MatrixSymbol("x", self.num_mapped_coords, 1)
        self.x = x
        self.N = Matrix(eval(mesh_data["shape_functions"]))
        self._compute_jacobian()

    def _compute_jacobian(self):

        assert self.num_vertices == len(self.N)
        assert self.num_dim == self.num_mapped_coords

        X = MatrixSymbol("vertices", self.num_vertices, self.num_dim)
        self.fx = MatrixSymbol("fx", self.num_dim, 1)
        physical_position = MatrixSymbol("phys_x", self.num_dim, 1)

        self.f = (self.N.T * Matrix(X)).T - physical_position

        self.J = symarray("J", (self.num_dim, self.num_dim))
        for i in range(self.num_dim):
            for j, var in enumerate(self.x):
                self.J[i][j] = diff(self.f[i, 0], var)

        self.rcol = MatrixSymbol("rcol", self.num_dim, 1)
        self.scol = MatrixSymbol("scol", self.num_dim, 1)
        self.tcol = MatrixSymbol("tcol", self.num_dim, 1)

        self.function_name = "%sFunction%dD" % (self.mesh_type, self.num_dim)
        self.function_header = fun_def_template % self.function_name
        self.function_declaration = fun_dec_template % self.function_name

        self.jacobian_name = "%sJacobian%dD" % (self.mesh_type, self.num_dim)

        if self.num_dim == 3:
            self.jacobian_header = jac_def_template_3D % self.jacobian_name
            self.jacobian_declaration = jac_dec_template_3D % self.jacobian_name

        elif self.num_dim == 2:
            self.jacobian_header = jac_def_template_2D % self.jacobian_name
            self.jacobian_declaration = jac_dec_template_2D % self.jacobian_name

    def get_interpolator_definition(self):
        """

        This returns the function definitions for the given mesh type.

        """

        function_code = self.function_header
        for i in range(self.num_dim):
            function_code += "\t" + ccode(self.f[i, 0], self.fx[i, 0]) + "\n"

        jacobian_code = self.jacobian_header
        for i in range(self.num_dim):
            jacobian_code += "\t" + ccode(self.J[i, 0], self.rcol[i, 0]) + "\n"
            jacobian_code += "\t" + ccode(self.J[i, 1], self.scol[i, 0]) + "\n"
            if self.num_dim == 2:
                continue
            jacobian_code += "\t" + ccode(self.J[i, 2], self.tcol[i, 0]) + "\n"

        return function_code, jacobian_code

    def get_interpolator_declaration(self):
        """

        This returns the function declarations for the given mesh type.

        """
        return self.function_declaration, self.jacobian_declaration


if __name__ == "__main__":

    with open("mesh_types.yaml") as f:
        lines = f.read()

    mesh_types = yaml.load(lines, Loader=yaml.FullLoader)

    pxd_file = open("lib/autogenerated_element_samplers.pxd", "w")
    pyx_file = open("lib/autogenerated_element_samplers.pyx", "w")

    pyx_file.write(file_header)
    pyx_file.write("\n \n")
    pyx_file.write("cimport cython \n")
    pyx_file.write("from libc.math cimport pow \n")
    pyx_file.write("\n \n")

    for _, mesh_data in sorted(mesh_types.items()):
        codegen = MeshCodeGenerator(mesh_data)

        function_code, jacobian_code = codegen.get_interpolator_definition()
        function_decl, jacobian_decl = codegen.get_interpolator_declaration()

        pxd_file.write(function_decl)
        pxd_file.write("\n \n")
        pxd_file.write(jacobian_decl)
        pxd_file.write("\n \n")

        pyx_file.write(function_code)
        pyx_file.write("\n \n")
        pyx_file.write(jacobian_code)
        pyx_file.write("\n \n")

    pxd_file.close()
    pyx_file.close()



import os
import sys

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.fields.field_info_container import NullFunc
from yt.frontends.enzo.misc import cosmology_get_units
from yt.frontends.enzo_e.fields import EnzoEFieldInfo
from yt.frontends.enzo_e.misc import (
    get_block_info,
    get_child_index,
    get_listed_subparam,
    get_root_block_id,
    get_root_blocks,
    is_parent,
    nested_dict_get,
)
from yt.funcs import get_pbar, setdefaultattr
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.cosmology import Cosmology
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py, _libconf as libconf

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from yt._maintenance.backports import cached_property


class EnzoEGrid(AMRGridPatch):
    """
    Class representing a single EnzoE Grid instance.
    """

    _id_offset = 0
    _refine_by = 2

    def __init__(self, id, index, block_name, filename=None):
        """
        Returns an instance of EnzoEGrid with *id*, associated with
        *filename* and *index*.
        """
        # All of the field parameters will be passed to us as needed.
        AMRGridPatch.__init__(self, id, filename=filename, index=index)
        self.block_name = block_name
        self._children_ids = None
        self._parent_id = -1
        self.Level = -1

    def __repr__(self):
        return "EnzoEGrid_%04d" % self.id

    def _prepare_grid(self):
        """Copies all the appropriate attributes from the index."""
        h = self.index  # cache it
        my_ind = self.id - self._id_offset
        self.ActiveDimensions = h.grid_dimensions[my_ind]
        self.LeftEdge = h.grid_left_edge[my_ind]
        self.RightEdge = h.grid_right_edge[my_ind]

    def get_parent_id(self, desc_block_name):
        if self.block_name == desc_block_name:
            raise RuntimeError("Child and parent are the same!")
        dim = self.ds.dimensionality
        d_block = desc_block_name[1:].replace(":", "")
        parent = self

        while True:
            a_block = parent.block_name[1:].replace(":", "")
            gengap = (len(d_block) - len(a_block)) / dim
            if gengap <= 1:
                return parent.id
            cid = get_child_index(a_block, d_block)
            parent = self.index.grids[parent._children_ids[cid]]

    def add_child(self, child):
        if self._children_ids is None:
            self._children_ids = -1 * np.ones(
                self._refine_by**self.ds.dimensionality, dtype=np.int64
            )

        a_block = self.block_name[1:].replace(":", "")
        d_block = child.block_name[1:].replace(":", "")
        cid = get_child_index(a_block, d_block)
        self._children_ids[cid] = child.id

    @cached_property
    def particle_count(self):
        with h5py.File(self.filename, mode="r") as f:
            fnstr = "{}/{}".format(
                self.block_name,
                self.ds.index.io._sep.join(["particle", "%s", "%s"]),
            )
            return {
                ptype: f.get(fnstr % (ptype, pfield)).size
                for ptype, pfield in self.ds.index.io.sample_pfields.items()
            }

    @cached_property
    def total_particles(self) -> int:
        return sum(self.particle_count.values())

    @property
    def Parent(self):
        if self._parent_id == -1:
            return None
        return self.index.grids[self._parent_id]

    @property
    def Children(self):
        if self._children_ids is None:
            return []
        return [self.index.grids[cid] for cid in self._children_ids]


class EnzoEHierarchy(GridIndex):

    _strip_path = False
    grid = EnzoEGrid
    _preload_implemented = True

    def __init__(self, ds, dataset_type):

        self.dataset_type = dataset_type
        self.directory = os.path.dirname(ds.parameter_filename)
        self.index_filename = ds.parameter_filename
        if os.path.getsize(self.index_filename) == 0:
            raise OSError(-1, "File empty", self.index_filename)

        GridIndex.__init__(self, ds, dataset_type)
        self.dataset.dataset_type = self.dataset_type

    def _count_grids(self):
        fblock_size = 32768
        f = open(self.ds.parameter_filename)
        f.seek(0, 2)
        file_size = f.tell()
        nblocks = np.ceil(float(file_size) / fblock_size).astype(np.int64)
        f.seek(0)
        offset = f.tell()
        ngrids = 0
        for _ in range(nblocks):
            my_block = min(fblock_size, file_size - offset)
            buff = f.read(my_block)
            ngrids += buff.count("\n")
            offset += my_block
        f.close()
        self.num_grids = ngrids
        self.dataset_type = "enzo_e"

    def _parse_index(self):
        self.grids = np.empty(self.num_grids, dtype="object")

        c = 1
        pbar = get_pbar("Parsing Hierarchy", self.num_grids)
        f = open(self.ds.parameter_filename)
        fblock_size = 32768
        f.seek(0, 2)
        file_size = f.tell()
        nblocks = np.ceil(float(file_size) / fblock_size).astype(np.int64)
        f.seek(0)
        offset = f.tell()
        lstr = ""
        # place child blocks after the root blocks
        rbdim = self.ds.root_block_dimensions
        nroot_blocks = rbdim.prod()
        child_id = nroot_blocks

        last_pid = None
        for _ib in range(nblocks):
            fblock = min(fblock_size, file_size - offset)
            buff = lstr + f.read(fblock)
            bnl = 0
            for _inl in range(buff.count("\n")):
                nnl = buff.find("\n", bnl)
                line = buff[bnl:nnl]
                block_name, block_file = line.split()

                # Handling of the B, B_, and B__ blocks is consistent with
                # other unrefined blocks
                level, left, right = get_block_info(block_name)
                rbindex = get_root_block_id(block_name)
                rbid = (
                    rbindex[0] * rbdim[1:].prod()
                    + rbindex[1] * rbdim[2:].prod()
                    + rbindex[2]
                )

                # There are also blocks at lower level than the
                # real root blocks. These can be ignored.
                if level == 0:
                    check_root = get_root_blocks(block_name).prod()
                    if check_root < nroot_blocks:
                        level = -1

                if level == -1:
                    grid_id = child_id
                    parent_id = -1
                    child_id += 1
                elif level == 0:
                    grid_id = rbid
                    parent_id = -1
                else:
                    grid_id = child_id
                    # Try the last parent_id first
                    if last_pid is not None and is_parent(
                        self.grids[last_pid].block_name, block_name
                    ):
                        parent_id = last_pid
                    else:
                        parent_id = self.grids[rbid].get_parent_id(block_name)
                    last_pid = parent_id
                    child_id += 1

                my_grid = self.grid(
                    grid_id,
                    self,
                    block_name,
                    filename=os.path.join(self.directory, block_file),
                )
                my_grid.Level = level
                my_grid._parent_id = parent_id

                self.grids[grid_id] = my_grid
                self.grid_levels[grid_id] = level
                self.grid_left_edge[grid_id] = left
                self.grid_right_edge[grid_id] = right
                self.grid_dimensions[grid_id] = self.ds.active_grid_dimensions

                if level > 0:
                    self.grids[parent_id].add_child(my_grid)

                bnl = nnl + 1
                pbar.update(c)
                c += 1
            lstr = buff[bnl:]
            offset += fblock

        f.close()
        pbar.finish()

        slope = self.ds.domain_width / self.ds.arr(np.ones(3), "code_length")
        self.grid_left_edge = self.grid_left_edge * slope + self.ds.domain_left_edge
        self.grid_right_edge = self.grid_right_edge * slope + self.ds.domain_left_edge

    def _populate_grid_objects(self):
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()
        self.max_level = self.grid_levels.max()

    def _setup_derived_fields(self):
        super()._setup_derived_fields()
        for fname, field in self.ds.field_info.items():
            if not field.particle_type:
                continue
            if isinstance(fname, tuple):
                continue
            if field._function is NullFunc:
                continue

    def _get_particle_type_counts(self):
        return {
            ptype: sum(g.particle_count[ptype] for g in self.grids)
            for ptype in self.ds.particle_types_raw
        }

    def _detect_output_fields(self):
        self.field_list = []
        # Do this only on the root processor to save disk work.
        if self.comm.rank in (0, None):
            # Just check the first grid.
            grid = self.grids[0]
            field_list, ptypes = self.io._read_field_names(grid)
            mylog.debug("Grid %s has: %s", grid.id, field_list)
            sample_pfields = self.io.sample_pfields
        else:
            field_list = None
            ptypes = None
            sample_pfields = None
        self.field_list = list(self.comm.mpi_bcast(field_list))
        self.dataset.particle_types = list(self.comm.mpi_bcast(ptypes))
        self.dataset.particle_types_raw = self.dataset.particle_types[:]
        self.io.sample_pfields = self.comm.mpi_bcast(sample_pfields)


class EnzoEDataset(Dataset):
    """
    Enzo-E-specific output, set at a fixed time.
    """

    refine_by = 2
    _index_class = EnzoEHierarchy
    _field_info_class = EnzoEFieldInfo
    _suffix = ".block_list"
    particle_types = None
    particle_types_raw = None

    def __init__(
        self,
        filename,
        dataset_type=None,
        parameter_override=None,
        conversion_override=None,
        storage_filename=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        """
        This class is a stripped down class that simply reads and parses
        *filename* without looking at the index.  *dataset_type* gets passed
        to the index to pre-determine the style of data-output.  However,
        it is not strictly necessary.  Optionally you may specify a
        *parameter_override* dictionary that will override anything in the
        parameter file and a *conversion_override* dictionary that consists
        of {fieldname : conversion_to_cgs} that will override the #DataCGS.
        """
        self.fluid_types += ("enzoe",)
        if parameter_override is None:
            parameter_override = {}
        self._parameter_override = parameter_override
        if conversion_override is None:
            conversion_override = {}
        self._conversion_override = conversion_override
        self.storage_filename = storage_filename
        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

    def _parse_parameter_file(self):
        """
        Parses the parameter file and establishes the various
        dictionaries.
        """

        f = open(self.parameter_filename)
        # get dimension from first block name
        b0, fn0 = f.readline().strip().split()
        level0, left0, right0 = get_block_info(b0, min_dim=0)
        root_blocks = get_root_blocks(b0)
        f.close()
        self.dimensionality = left0.size
        self._periodicity = tuple(np.ones(self.dimensionality, dtype=bool))

        lcfn = self.parameter_filename[: -len(self._suffix)] + ".libconfig"
        if os.path.exists(lcfn):
            with open(lcfn) as lf:
                self.parameters = libconf.load(lf)

            # Enzo-E ignores all cosmology parameters if "cosmology" is not in
            # the Physics:list parameter
            physics_list = nested_dict_get(
                self.parameters, ("Physics", "list"), default=[]
            )
            if "cosmology" in physics_list:
                self.cosmological_simulation = 1
                co_pars = [
                    "hubble_constant_now",
                    "omega_matter_now",
                    "omega_lambda_now",
                    "comoving_box_size",
                    "initial_redshift",
                ]
                co_dict = {
                    attr: nested_dict_get(
                        self.parameters, ("Physics", "cosmology", attr)
                    )
                    for attr in co_pars
                }
                for attr in ["hubble_constant", "omega_matter", "omega_lambda"]:
                    setattr(self, attr, co_dict[f"{attr}_now"])

                # Current redshift is not stored, so it's not possible
                # to set all cosmological units yet.
                # Get the time units and use that to figure out redshift.
                k = cosmology_get_units(
                    self.hubble_constant,
                    self.omega_matter,
                    co_dict["comoving_box_size"],
                    co_dict["initial_redshift"],
                    0,
                )
                setdefaultattr(self, "time_unit", self.quan(k["utim"], "s"))
                co = Cosmology(
                    hubble_constant=self.hubble_constant,
                    omega_matter=self.omega_matter,
                    omega_lambda=self.omega_lambda,
                )
            else:
                self.cosmological_simulation = 0
        else:
            self.cosmological_simulation = 0

        fh = h5py.File(os.path.join(self.directory, fn0), "r")
        self.domain_left_edge = fh.attrs["lower"]
        self.domain_right_edge = fh.attrs["upper"]
        if "version" in fh.attrs:
            version = fh.attrs.get("version").tobytes().decode("ascii")
        else:
            version = None  # earliest recorded version is '0.9.0'
        self.parameters["version"] = version

        # all blocks are the same size
        ablock = fh[list(fh.keys())[0]]
        self.current_time = ablock.attrs["time"][0]
        self.parameters["current_cycle"] = ablock.attrs["cycle"][0]
        gsi = ablock.attrs["enzo_GridStartIndex"]
        gei = ablock.attrs["enzo_GridEndIndex"]
        self.ghost_zones = gsi[0]
        self.root_block_dimensions = root_blocks
        self.active_grid_dimensions = gei - gsi + 1
        self.grid_dimensions = ablock.attrs["enzo_GridDimension"]
        self.domain_dimensions = root_blocks * self.active_grid_dimensions
        fh.close()

        if self.cosmological_simulation:
            self.current_redshift = co.z_from_t(self.current_time * self.time_unit)

        self._periodicity += (False,) * (3 - self.dimensionality)
        self._parse_fluid_prop_params()

    def _parse_fluid_prop_params(self):
        """
        Parse the fluid properties.
        """

        fp_params = nested_dict_get(
            self.parameters, ("Physics", "fluid_props"), default=None
        )

        if fp_params is not None:
            # in newer versions of enzo-e, this data is specified in a
            # centralized parameter group called Physics:fluid_props
            # -  for internal reasons related to backwards compatability,
            #    treatment of this physics-group is somewhat special (compared
            #    to the cosmology group). The parameters in this group are
            #    honored even if Physics:list does not include "fluid_props"
            self.gamma = nested_dict_get(fp_params, ("eos", "gamma"))
            de_type = nested_dict_get(
                fp_params, ("dual_energy", "type"), default="disabled"
            )
            uses_de = de_type != "disabled"
        else:
            # in older versions, these parameters were more scattered
            self.gamma = nested_dict_get(self.parameters, ("Field", "gamma"))

            uses_de = False
            for method in ("ppm", "mhd_vlct"):
                subparams = get_listed_subparam(
                    self.parameters, "Method", method, default=None
                )
                if subparams is not None:
                    uses_de = subparams.get("dual_energy", False)
        self.parameters["uses_dual_energy"] = uses_de

    def _set_code_unit_attributes(self):
        if self.cosmological_simulation:
            box_size = self.parameters["Physics"]["cosmology"]["comoving_box_size"]
            k = cosmology_get_units(
                self.hubble_constant,
                self.omega_matter,
                box_size,
                self.parameters["Physics"]["cosmology"]["initial_redshift"],
                self.current_redshift,
            )
            # Now some CGS values
            setdefaultattr(self, "length_unit", self.quan(box_size, "Mpccm/h"))
            setdefaultattr(
                self,
                "mass_unit",
                self.quan(k["urho"], "g/cm**3") * (self.length_unit.in_cgs()) ** 3,
            )
            setdefaultattr(self, "velocity_unit", self.quan(k["uvel"], "cm/s"))
        else:
            p = self.parameters
            for d, u in zip(("length", "time"), ("cm", "s")):
                val = nested_dict_get(p, ("Units", d), default=1)
                setdefaultattr(self, f"{d}_unit", self.quan(val, u))
            mass = nested_dict_get(p, ("Units", "mass"))
            if mass is None:
                density = nested_dict_get(p, ("Units", "density"))
                if density is not None:
                    mass = density * self.length_unit**3
                else:
                    mass = 1
            setdefaultattr(self, "mass_unit", self.quan(mass, "g"))
            setdefaultattr(self, "velocity_unit", self.length_unit / self.time_unit)

        magnetic_unit = np.sqrt(
            4 * np.pi * self.mass_unit / (self.time_unit**2 * self.length_unit)
        )
        magnetic_unit = np.float64(magnetic_unit.in_cgs())
        setdefaultattr(self, "magnetic_unit", self.quan(magnetic_unit, "gauss"))

    def __str__(self):
        return self.basename[: -len(self._suffix)]

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        ddir = os.path.dirname(filename)
        if not filename.endswith(cls._suffix):
            return False
        try:
            with open(filename) as f:
                block, block_file = f.readline().strip().split()
                get_block_info(block)
                if not os.path.exists(os.path.join(ddir, block_file)):
                    return False
        except Exception:
            return False
        return True



# Some tests for the Cylindrical coordinates handler

import numpy as np

from yt import SlicePlot
from yt.testing import add_noise_fields, assert_almost_equal, assert_equal, fake_amr_ds
from yt.utilities.answer_testing.framework import GenericImageTest

# Our canonical tests are that we can access all of our fields and we can
# compute our volume correctly.


def test_cylindrical_coordinates():
    # We're going to load up a simple AMR grid and check its volume
    # calculations and path length calculations.
    ds = fake_amr_ds(geometry="cylindrical")
    axes = ["r", "z", "theta"]
    for i, axis in enumerate(axes):
        dd = ds.all_data()
        fi = ("index", axis)
        fd = ("index", f"d{axis}")
        ma = np.argmax(dd[fi])
        assert_equal(dd[fi][ma] + dd[fd][ma] / 2.0, ds.domain_right_edge[i].d)
        mi = np.argmin(dd[fi])
        assert_equal(dd[fi][mi] - dd[fd][mi] / 2.0, ds.domain_left_edge[i].d)
        assert_equal(dd[fd].max(), (ds.domain_width / ds.domain_dimensions)[i].d)
    assert_almost_equal(
        dd[("index", "cell_volume")].sum(dtype="float64"),
        np.pi * ds.domain_width[0] ** 2 * ds.domain_width[1],
    )
    assert_equal(dd["index", "path_element_r"], dd["index", "dr"])
    assert_equal(dd["index", "path_element_z"], dd["index", "dz"])
    assert_equal(
        dd["index", "path_element_theta"], dd["index", "r"] * dd["index", "dtheta"]
    )


def test_noise_plots():
    ds = fake_amr_ds(geometry="cylindrical")
    add_noise_fields(ds)

    def create_image(filename_prefix):
        fields = ["noise%d" % i for i in range(4)]

        p = SlicePlot(ds, "z", fields)
        p.save(f"{filename_prefix}_log")

        p.set_log("all", False)
        p.save(f"{filename_prefix}_lin")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_noise_plot_lin"
    test_noise_plots.__name__ = test.description
    yield test



import pytest

import yt
from yt._version import VersionTuple, _parse_to_version_info


@pytest.mark.parametrize(
    "version_str, expected",
    (
        ("4.1.0", VersionTuple(4, 1, 0, "final", 0)),
        ("6.2.5", VersionTuple(6, 2, 5, "final", 0)),
        ("4.1.dev0", VersionTuple(4, 1, 0, "alpha", 0)),
        ("4.1.0rc", VersionTuple(4, 1, 0, "candidate", 0)),
        ("4.1.0rc1", VersionTuple(4, 1, 0, "candidate", 1)),
        ("4.1.0rc2", VersionTuple(4, 1, 0, "candidate", 2)),
    ),
)
def test_parse_version_info(version_str, expected):
    actual = _parse_to_version_info(version_str)
    assert actual == expected


def test_version_tuple_comp():
    # exercise comparison with builtin tuples
    # comparison results do not matter for this test

    yt.version_info > (4,)  # noqa: B015
    yt.version_info > (4, 1)  # noqa: B015
    yt.version_info > (4, 1, 0)  # noqa: B015
    yt.version_info < (4, 1, 0)  # noqa: B015
    yt.version_info <= (4, 1, 0)  # noqa: B015
    yt.version_info >= (4, 1, 0)  # noqa: B015
    yt.version_info == (4, 1, 0)  # noqa: B015

    assert isinstance(yt.version_info, tuple)



from . import tests
from .data_structures import (
    ARTDataset,
    ARTDomainFile,
    ARTDomainSubset,
    ARTIndex,
    DarkMatterARTDataset,
)
from .fields import ARTFieldInfo
from .io import IOHandlerART



import os
from copy import deepcopy
from pathlib import Path

from yt.frontends.amrvac.api import read_amrvac_namelist
from yt.testing import requires_module
from yt.utilities.on_demand_imports import _f90nml as f90nml

test_dir = os.path.dirname(os.path.abspath(__file__))
blast_wave_parfile = os.path.join(test_dir, "sample_parfiles", "bw_3d.par")
modifier_parfile = os.path.join(test_dir, "sample_parfiles", "tvdlf_scheme.par")


@requires_module("f90nml")
def test_read_pathlike():
    read_amrvac_namelist(Path(blast_wave_parfile))


@requires_module("f90nml")
def test_read_one_file():
    """when provided a single file, the function should merely act
    as a wrapper for f90nml.read()"""
    namelist1 = read_amrvac_namelist(blast_wave_parfile)
    namelist2 = f90nml.read(blast_wave_parfile)
    assert namelist1 == namelist2


@requires_module("f90nml")
def test_accumulate_basename():
    """When two (or more) parfiles are passed,
    the filelist:base_filename should be special-cased"""
    namelist_base = f90nml.read(blast_wave_parfile)
    namelist_update = f90nml.read(modifier_parfile)

    namelist_tot1 = read_amrvac_namelist([blast_wave_parfile, modifier_parfile])
    namelist_tot2 = deepcopy(namelist_base)
    namelist_tot2.patch(namelist_update)

    # remove and store the special-case value
    name1 = namelist_tot1["filelist"].pop("base_filename")
    name2 = namelist_tot2["filelist"].pop("base_filename")

    assert (
        name1
        == namelist_base["filelist"]["base_filename"]
        + namelist_update["filelist"]["base_filename"]
    )
    assert name2 == namelist_update["filelist"]["base_filename"]
    assert name1 != name2

    # test equality for the rest of the namelist
    assert namelist_tot1 == namelist_tot2



import numpy as np


def enhance(im, stdval=6.0, just_alpha=True):
    if just_alpha:
        nz = im[im > 0.0]
        im[:] = im[:] / (nz.mean() + stdval * np.std(nz))
    else:
        for c in range(3):
            nz = im[:, :, c][im[:, :, c] > 0.0]
            im[:, :, c] = im[:, :, c] / (nz.mean() + stdval * np.std(nz))
            del nz
    np.clip(im, 0.0, 1.0, im)


def enhance_rgba(im, stdval=6.0):
    nzc = im[:, :, :3][im[:, :, :3] > 0.0]
    cmax = nzc.mean() + stdval * nzc.std()

    nza = im[:, :, 3][im[:, :, 3] > 0.0]
    if len(nza) == 0:
        im[:, :, 3] = 1.0
        amax = 1.0
    else:
        amax = nza.mean() + stdval * nza.std()

    im.rescale(amax=amax, cmax=cmax, inline=True)
    np.clip(im, 0.0, 1.0, im)



import glob
import os

import numpy as np

from yt.data_objects.static_output import ParticleDataset
from yt.frontends.halo_catalog.data_structures import HaloCatalogFile
from yt.funcs import setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities import fortran_utils as fpu
from yt.utilities.cosmology import Cosmology

from .definitions import header_dt
from .fields import RockstarFieldInfo


class RockstarBinaryFile(HaloCatalogFile):
    def __init__(self, ds, io, filename, file_id, range):
        with open(filename, "rb") as f:
            self.header = fpu.read_cattrs(f, header_dt, "=")
            self._position_offset = f.tell()
            f.seek(0, os.SEEK_END)
            self._file_size = f.tell()

        super().__init__(ds, io, filename, file_id, range)

    def _read_particle_positions(self, ptype, f=None):
        """
        Read all particle positions in this file.
        """

        if f is None:
            close = True
            f = open(self.filename, "rb")
        else:
            close = False

        pcount = self.header["num_halos"]
        pos = np.empty((pcount, 3), dtype="float64")
        f.seek(self._position_offset, os.SEEK_SET)
        halos = np.fromfile(f, dtype=self.io._halo_dt, count=pcount)
        for i, ax in enumerate("xyz"):
            pos[:, i] = halos[f"particle_position_{ax}"].astype("float64")

        if close:
            f.close()

        return pos


class RockstarDataset(ParticleDataset):
    _index_class = ParticleIndex
    _file_class = RockstarBinaryFile
    _field_info_class = RockstarFieldInfo
    _suffix = ".bin"

    def __init__(
        self,
        filename,
        dataset_type="rockstar_binary",
        units_override=None,
        unit_system="cgs",
        index_order=None,
        index_filename=None,
    ):
        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _parse_parameter_file(self):
        with open(self.parameter_filename, "rb") as f:
            hvals = fpu.read_cattrs(f, header_dt)
            hvals.pop("unused")
        self.dimensionality = 3
        self.refine_by = 2
        prefix = ".".join(self.parameter_filename.rsplit(".", 2)[:-2])
        self.filename_template = f"{prefix}.%(num)s{self._suffix}"
        self.file_count = len(glob.glob(prefix + ".*" + self._suffix))

        # Now we can set up things we already know.
        self.cosmological_simulation = 1
        self.current_redshift = (1.0 / hvals["scale"]) - 1.0
        self.hubble_constant = hvals["h0"]
        self.omega_lambda = hvals["Ol"]
        self.omega_matter = hvals["Om"]
        cosmo = Cosmology(
            hubble_constant=self.hubble_constant,
            omega_matter=self.omega_matter,
            omega_lambda=self.omega_lambda,
        )
        self.current_time = cosmo.lookback_time(self.current_redshift, 1e6).in_units(
            "s"
        )
        self._periodicity = (True, True, True)
        self.particle_types = "halos"
        self.particle_types_raw = "halos"

        self.domain_left_edge = np.array([0.0, 0.0, 0.0])
        self.domain_right_edge = np.array([hvals["box_size"]] * 3)

        self.domain_dimensions = np.ones(3, "int32")
        self.parameters.update(hvals)

    def _set_code_unit_attributes(self):
        z = self.current_redshift
        setdefaultattr(self, "length_unit", self.quan(1.0 / (1.0 + z), "Mpc / h"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "Msun / h"))
        setdefaultattr(self, "velocity_unit", self.quan(1.0, "km / s"))
        setdefaultattr(self, "time_unit", self.length_unit / self.velocity_unit)

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".bin"):
            return False
        with open(filename, mode="rb") as f:
            header = fpu.read_cattrs(f, header_dt)
            if header["magic"] == 18077126535843729616:
                return True
        return False



from yt.utilities.lib.cykdtree import plot  # NOQA
from yt.utilities.lib.cykdtree.kdtree import PyKDTree, PyNode  # NOQA


def make_tree(pts, **kwargs):
    r"""Build a KD-tree for a set of points.

    Args:
        pts (np.ndarray of float64): (n,m) Array of n mD points.
        \*\*kwargs: Additional keyword arguments are passed to the appropriate
            class for constructuing the tree.

    Returns:
        T (:class:`cykdtree.PyKDTree`): KDTree object.

    Raises:
        ValueError: If `pts` is not a 2D array.

    """
    # Check input
    if pts.ndim != 2:
        raise ValueError("pts must be a 2D array of ND coordinates")
    T = PyKDTree(pts, **kwargs)
    return T



from itertools import product

import numpy as np

from yt.frontends.open_pmd.data_structures import OpenPMDDataset
from yt.loaders import load
from yt.testing import (
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
    requires_file,
)
from yt.utilities.answer_testing.framework import data_dir_load

twoD = "example-2d/hdf5/data00000100.h5"
threeD = "example-3d/hdf5/data00000100.h5"
noFields = "no_fields/data00000400.h5"
noParticles = "no_particles/data00000400.h5"
groupBased = "singleParticle/simData.h5"

particle_fields = [
    "particle_charge",
    "particle_mass",
    "particle_momentum_x",
    "particle_momentum_y",
    "particle_momentum_z",
    "particle_positionCoarse_x",
    "particle_positionCoarse_y",
    "particle_positionCoarse_z",
    "particle_positionOffset_x",
    "particle_positionOffset_y",
    "particle_positionOffset_z",
    "particle_weighting",
]


@requires_file(threeD)
def test_3d_out():
    ds = data_dir_load(threeD)
    particle_types = ["all", "io", "nbody"]
    field_list = list(product(particle_types, particle_fields))
    field_list += list(product(("openPMD",), ("E_x", "E_y", "E_z", "rho")))
    domain_dimensions = [26, 26, 201] * np.ones_like(ds.domain_dimensions)
    domain_width = [2.08e-05, 2.08e-05, 2.01e-05] * np.ones_like(ds.domain_left_edge)

    assert isinstance(ds, OpenPMDDataset)
    assert_equal(str(ds), "data00000100.h5")
    assert_equal(ds.dimensionality, 3)
    assert_equal(ds.particle_types_raw, ("io",))
    assert "all" in ds.particle_unions
    assert_array_equal(ds.field_list, field_list)
    assert_array_equal(ds.domain_dimensions, domain_dimensions)
    assert_almost_equal(
        ds.current_time, 3.28471214521e-14 * np.ones_like(ds.current_time)
    )
    assert_almost_equal(ds.domain_right_edge - ds.domain_left_edge, domain_width)


@requires_file(twoD)
def test_2d_out():
    ds = data_dir_load(twoD)
    particle_types = ("Hydrogen1+", "all", "electrons", "nbody")
    field_list = list(product(particle_types, particle_fields))
    field_list += [
        ("openPMD", "B_x"),
        ("openPMD", "B_y"),
        ("openPMD", "B_z"),
        ("openPMD", "E_x"),
        ("openPMD", "E_y"),
        ("openPMD", "E_z"),
        ("openPMD", "J_x"),
        ("openPMD", "J_y"),
        ("openPMD", "J_z"),
        ("openPMD", "rho"),
    ]
    domain_dimensions = [51, 201, 1] * np.ones_like(ds.domain_dimensions)
    domain_width = [3.06e-05, 2.01e-05, 1e0] * np.ones_like(ds.domain_left_edge)

    assert isinstance(ds, OpenPMDDataset)
    assert_equal(str(ds), "data00000100.h5")
    assert_equal(ds.dimensionality, 2)
    assert_equal(ds.particle_types_raw, ("Hydrogen1+", "electrons"))
    assert "all" in ds.particle_unions
    assert_array_equal(ds.field_list, field_list)
    assert_array_equal(ds.domain_dimensions, domain_dimensions)
    assert_almost_equal(
        ds.current_time, 3.29025596712e-14 * np.ones_like(ds.current_time)
    )
    assert_almost_equal(ds.domain_right_edge - ds.domain_left_edge, domain_width)


@requires_file(noFields)
def test_no_fields_out():
    ds = data_dir_load(noFields)
    particle_types = ("all", "io", "nbody")
    no_fields_pfields = sorted(particle_fields + ["particle_id"])
    field_list = list(product(particle_types, no_fields_pfields))
    domain_dimensions = [1, 1, 1] * np.ones_like(ds.domain_dimensions)
    domain_width = [1, 1, 1] * np.ones_like(ds.domain_left_edge)

    assert isinstance(ds, OpenPMDDataset)
    assert_equal(str(ds), "data00000400.h5")
    assert_equal(ds.dimensionality, 3)
    assert_equal(ds.particle_types_raw, ("io",))
    assert "all" in ds.particle_unions
    assert_array_equal(ds.field_list, field_list)
    assert_array_equal(ds.domain_dimensions, domain_dimensions)
    assert_almost_equal(
        ds.current_time, 1.3161023868481013e-13 * np.ones_like(ds.current_time)
    )
    assert_almost_equal(ds.domain_right_edge - ds.domain_left_edge, domain_width)


@requires_file(noParticles)
def test_no_particles_out():
    ds = data_dir_load(noParticles)
    field_list = [
        ("openPMD", "E_x"),
        ("openPMD", "E_y"),
        ("openPMD", "E_z"),
        ("openPMD", "rho"),
    ]
    domain_dimensions = [51, 201, 1] * np.ones_like(ds.domain_dimensions)
    domain_width = [3.06e-05, 2.01e-05, 1e0] * np.ones_like(ds.domain_left_edge)

    assert isinstance(ds, OpenPMDDataset)
    assert_equal(str(ds), "data00000400.h5")
    assert_equal(ds.dimensionality, 2)
    assert_equal(ds.particle_types_raw, ("io",))
    assert "all" not in ds.particle_unions
    assert_array_equal(ds.field_list, field_list)
    assert_array_equal(ds.domain_dimensions, domain_dimensions)
    assert_almost_equal(
        ds.current_time, 1.3161023868481013e-13 * np.ones_like(ds.current_time)
    )
    assert_almost_equal(ds.domain_right_edge - ds.domain_left_edge, domain_width)


@requires_file(groupBased)
def test_groupBased_out():
    dss = load(groupBased)
    particle_types = ("all", "io", "nbody")
    field_list = list(product(particle_types, particle_fields))
    field_list += [
        ("openPMD", "J_x"),
        ("openPMD", "J_y"),
        ("openPMD", "J_z"),
        ("openPMD", "e-chargeDensity"),
    ]
    domain_dimensions = [32, 64, 64] * np.ones_like(dss[0].domain_dimensions)
    domain_width = [0.0002752, 0.0005504, 0.0005504] * np.ones_like(
        dss[0].domain_left_edge
    )

    assert_equal(len(dss), 101)
    for i in range(0, len(dss), 20):  # Test only every 20th ds out of the series
        ds = dss[i]
        assert_equal(str(ds), "simData.h5")
        assert_equal(ds.dimensionality, 3)
        assert_equal(ds.particle_types_raw, ("io",))
        assert_array_equal(ds.field_list, field_list)
        assert_array_equal(ds.domain_dimensions, domain_dimensions)
        assert ds.current_time >= np.zeros_like(ds.current_time)
        assert ds.current_time <= 1.6499999999999998e-12 * np.ones_like(ds.current_time)
        assert_almost_equal(ds.domain_right_edge - ds.domain_left_edge, domain_width)



cdef class AlwaysSelector(SelectorObject):

    def __init__(self, dobj):
        self.overlap_cells = 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        cdef int ng = left_edges.shape[0]
        cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.ones(ng, dtype='uint8')
        return gridi.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
        return 1

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) nogil:
        return 1

    cdef int select_point(self, np.float64_t pos[3]) nogil:
        return 1

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
        return 1

    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        return 1

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        return 1

    def _hash_vals(self):
        return ("always", 1,)

always_selector = AlwaysSelector



import numpy as np
from more_itertools import collapse

from yt.data_objects.field_data import YTFieldData
from yt.fields.derived_field import DerivedField
from yt.frontends.ytdata.utilities import save_as_dataset
from yt.funcs import get_output_filename, is_sequence, iter_fields, mylog
from yt.units.unit_object import Unit  # type: ignore
from yt.units.yt_array import YTQuantity, array_like_field
from yt.utilities.exceptions import (
    YTIllDefinedBounds,
    YTIllDefinedProfile,
    YTProfileDataShape,
)
from yt.utilities.lib.misc_utilities import (
    new_bin_profile1d,
    new_bin_profile2d,
    new_bin_profile3d,
)
from yt.utilities.lib.particle_mesh_operations import CICDeposit_2, NGPDeposit_2
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
    parallel_objects,
)


def _sanitize_min_max_units(amin, amax, finfo, registry):
    # returns a copy of amin and amax, converted to finfo's output units
    umin = getattr(amin, "units", None)
    umax = getattr(amax, "units", None)
    if umin is None:
        umin = Unit(finfo.output_units, registry=registry)
        rmin = YTQuantity(amin, umin)
    else:
        rmin = amin.in_units(finfo.output_units)
    if umax is None:
        umax = Unit(finfo.output_units, registry=registry)
        rmax = YTQuantity(amax, umax)
    else:
        rmax = amax.in_units(finfo.output_units)
    return rmin, rmax


def preserve_source_parameters(func):
    def save_state(*args, **kwargs):
        # Temporarily replace the 'field_parameters' for a
        # grid with the 'field_parameters' for the data source
        prof = args[0]
        source = args[1]
        if hasattr(source, "field_parameters"):
            old_params = source.field_parameters
            source.field_parameters = prof._data_source.field_parameters
            tr = func(*args, **kwargs)
            source.field_parameters = old_params
        else:
            tr = func(*args, **kwargs)
        return tr

    return save_state


class ProfileFieldAccumulator:
    def __init__(self, n_fields, size):
        shape = size + (n_fields,)
        self.values = np.zeros(shape, dtype="float64")
        self.mvalues = np.zeros(shape, dtype="float64")
        self.qvalues = np.zeros(shape, dtype="float64")
        self.used = np.zeros(size, dtype="bool")
        self.weight_values = np.zeros(size, dtype="float64")


class ProfileND(ParallelAnalysisInterface):
    """The profile object class"""

    def __init__(self, data_source, weight_field=None):
        self.data_source = data_source
        self.ds = data_source.ds
        self.field_map = {}
        self.field_info = {}
        self.field_data = YTFieldData()
        if weight_field is not None:
            self.standard_deviation = YTFieldData()
            weight_field = self.data_source._determine_fields(weight_field)[0]
        else:
            self.standard_deviation = None
        self.weight_field = weight_field
        self.field_units = {}
        ParallelAnalysisInterface.__init__(self, comm=data_source.comm)

    def add_fields(self, fields):
        """Add fields to profile

        Parameters
        ----------
        fields : list of field names
            A list of fields to create profile histograms for

        """
        fields = self.data_source._determine_fields(fields)
        for f in fields:
            self.field_info[f] = self.data_source.ds.field_info[f]
        temp_storage = ProfileFieldAccumulator(len(fields), self.size)
        citer = self.data_source.chunks([], "io")
        for chunk in parallel_objects(citer):
            self._bin_chunk(chunk, fields, temp_storage)
        self._finalize_storage(fields, temp_storage)

    def set_field_unit(self, field, new_unit):
        """Sets a new unit for the requested field

        Parameters
        ----------
        field : string or field tuple
           The name of the field that is to be changed.

        new_unit : string or Unit object
           The name of the new unit.
        """
        if field in self.field_units:
            self.field_units[field] = Unit(new_unit, registry=self.ds.unit_registry)
        else:
            fd = self.field_map[field]
            if fd in self.field_units:
                self.field_units[fd] = Unit(new_unit, registry=self.ds.unit_registry)
            else:
                raise KeyError(f"{field} not in profile!")

    def _finalize_storage(self, fields, temp_storage):
        # We use our main comm here
        # This also will fill _field_data

        for i, _field in enumerate(fields):
            # q values are returned as q * weight but we want just q
            temp_storage.qvalues[..., i][
                temp_storage.used
            ] /= temp_storage.weight_values[temp_storage.used]

        # get the profile data from all procs
        all_store = {self.comm.rank: temp_storage}
        all_store = self.comm.par_combine_object(all_store, "join", datatype="dict")

        all_val = np.zeros_like(temp_storage.values)
        all_mean = np.zeros_like(temp_storage.mvalues)
        all_std = np.zeros_like(temp_storage.qvalues)
        all_weight = np.zeros_like(temp_storage.weight_values)
        all_used = np.zeros_like(temp_storage.used, dtype="bool")

        # Combine the weighted mean and standard deviation from each processor.
        # For two samples with total weight, mean, and standard deviation
        # given by w, m, and s, their combined mean and standard deviation are:
        # m12 = (m1 * w1 + m2 * w2) / (w1 + w2)
        # s12 = (m1 * (s1**2 + (m1 - m12)**2) +
        #        m2 * (s2**2 + (m2 - m12)**2)) / (w1 + w2)
        # Here, the mvalues are m and the qvalues are s**2.
        for p in sorted(all_store.keys()):
            all_used += all_store[p].used
            old_mean = all_mean.copy()
            old_weight = all_weight.copy()
            all_weight[all_store[p].used] += all_store[p].weight_values[
                all_store[p].used
            ]
            for i, _field in enumerate(fields):
                all_val[..., i][all_store[p].used] += all_store[p].values[..., i][
                    all_store[p].used
                ]

                all_mean[..., i][all_store[p].used] = (
                    all_mean[..., i] * old_weight
                    + all_store[p].mvalues[..., i] * all_store[p].weight_values
                )[all_store[p].used] / all_weight[all_store[p].used]

                all_std[..., i][all_store[p].used] = (
                    old_weight
                    * (all_std[..., i] + (old_mean[..., i] - all_mean[..., i]) ** 2)
                    + all_store[p].weight_values
                    * (
                        all_store[p].qvalues[..., i]
                        + (all_store[p].mvalues[..., i] - all_mean[..., i]) ** 2
                    )
                )[all_store[p].used] / all_weight[all_store[p].used]

        all_std = np.sqrt(all_std)
        del all_store
        self.used = all_used
        blank = ~all_used

        self.weight = all_weight
        self.weight[blank] = 0.0

        for i, field in enumerate(fields):
            if self.weight_field is None:
                self.field_data[field] = array_like_field(
                    self.data_source, all_val[..., i], field
                )
            else:
                self.field_data[field] = array_like_field(
                    self.data_source, all_mean[..., i], field
                )
                self.standard_deviation[field] = array_like_field(
                    self.data_source, all_std[..., i], field
                )
                self.standard_deviation[field][blank] = 0.0
                self.weight = array_like_field(
                    self.data_source, self.weight, self.weight_field
                )
            self.field_data[field][blank] = 0.0
            self.field_units[field] = self.field_data[field].units
            if isinstance(field, tuple):
                self.field_map[field[1]] = field
            else:
                self.field_map[field] = field

    def _bin_chunk(self, chunk, fields, storage):
        raise NotImplementedError

    def _filter(self, bin_fields):
        # cut_points is set to be everything initially, but
        # we also want to apply a filtering based on min/max
        pfilter = np.ones(bin_fields[0].shape, dtype="bool")
        for (mi, ma), data in zip(self.bounds, bin_fields):
            pfilter &= data > mi
            pfilter &= data < ma
        return pfilter, [data[pfilter] for data in bin_fields]

    def _get_data(self, chunk, fields):
        # We are using chunks now, which will manage the field parameters and
        # the like.
        bin_fields = [chunk[bf] for bf in self.bin_fields]
        for i in range(1, len(bin_fields)):
            if bin_fields[0].shape != bin_fields[i].shape:
                raise YTProfileDataShape(
                    self.bin_fields[0],
                    bin_fields[0].shape,
                    self.bin_fields[i],
                    bin_fields[i].shape,
                )
        # We want to make sure that our fields are within the bounds of the
        # binning
        pfilter, bin_fields = self._filter(bin_fields)
        if not np.any(pfilter):
            return None
        arr = np.zeros((bin_fields[0].size, len(fields)), dtype="float64")
        for i, field in enumerate(fields):
            if pfilter.shape != chunk[field].shape:
                raise YTProfileDataShape(
                    self.bin_fields[0], bin_fields[0].shape, field, chunk[field].shape
                )
            units = chunk.ds.field_info[field].output_units
            arr[:, i] = chunk[field][pfilter].in_units(units)
        if self.weight_field is not None:
            if pfilter.shape != chunk[self.weight_field].shape:
                raise YTProfileDataShape(
                    self.bin_fields[0],
                    bin_fields[0].shape,
                    self.weight_field,
                    chunk[self.weight_field].shape,
                )
            units = chunk.ds.field_info[self.weight_field].output_units
            weight_data = chunk[self.weight_field].in_units(units)
        else:
            weight_data = np.ones(pfilter.shape, dtype="float64")
        weight_data = weight_data[pfilter]
        # So that we can pass these into
        return arr, weight_data, bin_fields

    def __getitem__(self, field):
        if field in self.field_data:
            fname = field
        else:
            # deal with string vs tuple field names and attempt to guess which field
            # we are supposed to be talking about
            fname = self.field_map.get(field, None)
            if isinstance(field, tuple):
                fname = self.field_map.get(field[1], None)
                if fname != field:
                    raise KeyError(
                        "Asked for field '{}' but only have data for "
                        "fields '{}'".format(field, list(self.field_data.keys()))
                    )
            elif isinstance(field, DerivedField):
                fname = self.field_map.get(field.name[1], None)
            if fname is None:
                raise KeyError(field)
        if getattr(self, "fractional", False):
            return self.field_data[fname]
        else:
            return self.field_data[fname].in_units(self.field_units[fname])

    def items(self):
        return [(k, self[k]) for k in self.field_data.keys()]

    def keys(self):
        return self.field_data.keys()

    def __iter__(self):
        return sorted(self.items())

    def _get_bins(self, mi, ma, n, take_log):
        if take_log:
            ret = np.logspace(np.log10(mi), np.log10(ma), n + 1)
            # at this point ret[0] and ret[-1] are not exactly equal to
            # mi and ma due to round-off error. Let's force them to be
            # mi and ma exactly to avoid incorrectly discarding cells near
            # the edges. See Issue #1300.
            ret[0], ret[-1] = mi, ma
            return ret
        else:
            return np.linspace(mi, ma, n + 1)

    def save_as_dataset(self, filename=None):
        r"""Export a profile to a reloadable yt dataset.

        This function will take a profile and output a dataset
        containing all relevant fields.  The resulting dataset
        can be reloaded as a yt dataset.

        Parameters
        ----------
        filename : str, optional
            The name of the file to be written.  If None, the name
            will be a combination of the original dataset plus
            the type of object, e.g., Profile1D.

        Returns
        -------
        filename : str
            The name of the file that has been created.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
        >>> ad = ds.all_data()
        >>> profile = yt.create_profile(
        ...     ad,
        ...     [("gas", "density"), ("gas", "temperature")],
        ...     ("gas", "mass"),
        ...     weight_field=None,
        ...     n_bins=(128, 128),
        ... )
        >>> fn = profile.save_as_dataset()
        >>> prof_ds = yt.load(fn)
        >>> print(prof_ds.data[("gas", "mass")])
        (128, 128)
        >>> print(prof_ds.data[("index", "x")].shape)  # x bins as 1D array
        (128,)
        >>> print(prof_ds.data[("gas", "density")])  # x bins as 2D array
        (128, 128)
        >>> p = yt.PhasePlot(
        ...     prof_ds.data,
        ...     ("gas", "density"),
        ...     ("gas", "temperature"),
        ...     ("gas", "mass"),
        ...     weight_field=None,
        ... )
        >>> p.save()

        """

        keyword = f"{str(self.ds)}_{self.__class__.__name__}"
        filename = get_output_filename(filename, keyword, ".h5")

        args = ("field", "log")
        extra_attrs = {
            "data_type": "yt_profile",
            "profile_dimensions": self.size,
            "weight_field": self.weight_field,
            "fractional": self.fractional,
            "accumulation": self.accumulation,
        }
        data = {}
        data.update(self.field_data)
        data["weight"] = self.weight
        data["used"] = self.used.astype("float64")
        std = "standard_deviation"
        if self.weight_field is not None:
            std_data = getattr(self, std)
            data.update({(std, field[1]): std_data[field] for field in self.field_data})

        dimensionality = 0
        bin_data = []
        for ax in "xyz":
            if hasattr(self, ax):
                dimensionality += 1
                data[ax] = getattr(self, ax)
                bin_data.append(data[ax])
                bin_field_name = f"{ax}_bins"
                data[bin_field_name] = getattr(self, bin_field_name)
                extra_attrs[f"{ax}_range"] = self.ds.arr(
                    [data[bin_field_name][0], data[bin_field_name][-1]]
                )
                for arg in args:
                    key = f"{ax}_{arg}"
                    extra_attrs[key] = getattr(self, key)

        bin_fields = np.meshgrid(*bin_data)
        for i, ax in enumerate("xyz"[:dimensionality]):
            data[getattr(self, f"{ax}_field")] = bin_fields[i]

        extra_attrs["dimensionality"] = dimensionality
        ftypes = {field: "data" for field in data if field[0] != std}
        if self.weight_field is not None:
            ftypes.update({(std, field[1]): std for field in self.field_data})
        save_as_dataset(
            self.ds, filename, data, field_types=ftypes, extra_attrs=extra_attrs
        )

        return filename


class ProfileNDFromDataset(ProfileND):
    """
    An ND profile object loaded from a ytdata dataset.
    """

    def __init__(self, ds):
        ProfileND.__init__(self, ds.data, ds.parameters.get("weight_field", None))
        self.fractional = ds.parameters.get("fractional", False)
        self.accumulation = ds.parameters.get("accumulation", False)
        exclude_fields = ["used", "weight"]
        for ax in "xyz"[: ds.dimensionality]:
            setattr(self, ax, ds.data[("data", ax)])
            ax_bins = f"{ax}_bins"
            ax_field = f"{ax}_field"
            ax_log = f"{ax}_log"
            setattr(self, ax_bins, ds.data[("data", ax_bins)])
            field_name = tuple(ds.parameters.get(ax_field, (None, None)))
            setattr(self, ax_field, field_name)
            self.field_info[field_name] = ds.field_info[field_name]
            setattr(self, ax_log, ds.parameters.get(ax_log, False))
            exclude_fields.extend([ax, ax_bins, field_name[1]])
        self.weight = ds.data[("data", "weight")]
        self.used = ds.data[("data", "used")].d.astype(bool)
        profile_fields = [
            f
            for f in ds.field_list
            if f[1] not in exclude_fields and f[0] != "standard_deviation"
        ]
        for field in profile_fields:
            self.field_map[field[1]] = field
            self.field_data[field] = ds.data[field]
            self.field_info[field] = ds.field_info[field]
            self.field_units[field] = ds.data[field].units
            if ("standard_deviation", field[1]) in ds.field_list:
                self.standard_deviation[field] = ds.data["standard_deviation", field[1]]


class Profile1D(ProfileND):
    """An object that represents a 1D profile.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_log : boolean
        Controls whether or not the bins for the x field are evenly
        spaced in linear (False) or log (True) space.
    weight_field : string field name
        The field to weight the profiled fields by.
    override_bins_x : array
        Array to set as xbins and ignore other parameters if set

    """

    def __init__(
        self,
        data_source,
        x_field,
        x_n,
        x_min,
        x_max,
        x_log,
        weight_field=None,
        override_bins_x=None,
    ):
        super().__init__(data_source, weight_field)
        self.x_field = data_source._determine_fields(x_field)[0]
        self.field_info[self.x_field] = self.data_source.ds.field_info[self.x_field]
        self.x_log = x_log
        x_min, x_max = _sanitize_min_max_units(
            x_min, x_max, self.field_info[self.x_field], self.ds.unit_registry
        )
        self.x_bins = array_like_field(
            data_source, self._get_bins(x_min, x_max, x_n, x_log), self.x_field
        )

        if override_bins_x is not None:
            self.x_bins = array_like_field(data_source, override_bins_x, self.x_field)

        self.size = (self.x_bins.size - 1,)
        self.bin_fields = (self.x_field,)
        self.x = 0.5 * (self.x_bins[1:] + self.x_bins[:-1])

    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None:
            return
        fdata, wdata, (bf_x,) = rv
        bf_x.convert_to_units(self.field_info[self.x_field].output_units)
        bin_ind = np.digitize(bf_x, self.x_bins) - 1
        new_bin_profile1d(
            bin_ind,
            wdata,
            fdata,
            storage.weight_values,
            storage.values,
            storage.mvalues,
            storage.qvalues,
            storage.used,
        )

        # We've binned it!

    def set_x_unit(self, new_unit):
        """Sets a new unit for the x field

        Parameters
        ----------
        new_unit : string or Unit object
            The name of the new unit.
        """
        self.x_bins.convert_to_units(new_unit)
        self.x = 0.5 * (self.x_bins[1:] + self.x_bins[:-1])

    @property
    def bounds(self):
        return ((self.x_bins[0], self.x_bins[-1]),)

    def plot(self):
        r"""
        This returns a :class:`~yt.visualization.profile_plotter.ProfilePlot`
        with the fields that have been added to this object.
        """
        from yt.visualization.profile_plotter import ProfilePlot

        return ProfilePlot.from_profiles(self)

    def _export_prep(self, fields, only_used):
        if only_used:
            idxs = self.used
        else:
            idxs = slice(None, None, None)
        if not only_used and not np.all(self.used):
            masked = True
        else:
            masked = False
        if fields is None:
            fields = self.field_data.keys()
        else:
            fields = self.data_source._determine_fields(fields)
        return idxs, masked, fields

    def to_dataframe(self, fields=None, only_used=False):
        r"""Export a profile object to a pandas DataFrame.

        This function will take a data object and construct from it and
        optionally a list of fields a pandas DataFrame object.  If pandas is
        not importable, this will raise ImportError.

        Parameters
        ----------
        fields : list of strings or tuple field names, default None
            If this is supplied, it is the list of fields to be exported into
            the DataFrame. If not supplied, whatever fields exist in the
            profile, along with the bin field, will be exported.
        only_used : boolean, default False
            If True, only the bins which have data will be exported. If False,
            all of the bins will be exported, but the elements for those bins
            in the data arrays will be filled with NaNs.

        Returns
        -------
        df : :class:`~pandas.DataFrame`
            The data contained in the profile.

        Examples
        --------
        >>> sp = ds.sphere("c", (0.1, "unitary"))
        >>> p = sp.profile(
        ...     ("index", "radius"), [("gas", "density"), ("gas", "temperature")]
        ... )
        >>> df1 = p.to_dataframe()
        >>> df2 = p.to_dataframe(fields=("gas", "density"), only_used=True)
        """
        from yt.utilities.on_demand_imports import _pandas as pd

        idxs, masked, fields = self._export_prep(fields, only_used)
        pdata = {self.x_field[-1]: self.x[idxs]}
        for field in fields:
            pdata[field[-1]] = self[field][idxs]
        df = pd.DataFrame(pdata)
        if masked:
            mask = np.zeros(df.shape, dtype="bool")
            mask[~self.used, 1:] = True
            df.mask(mask, inplace=True)
        return df

    def to_astropy_table(self, fields=None, only_used=False):
        """
        Export the profile data to a :class:`~astropy.table.table.QTable`,
        which is a Table object which is unit-aware. The QTable can then
        be exported to an ASCII file, FITS file, etc.

        See the AstroPy Table docs for more details:
        http://docs.astropy.org/en/stable/table/

        Parameters
        ----------
        fields : list of strings or tuple field names, default None
            If this is supplied, it is the list of fields to be exported into
            the DataFrame. If not supplied, whatever fields exist in the
            profile, along with the bin field, will be exported.
        only_used : boolean, optional
            If True, only the bins which are used are copied
            to the QTable as rows. If False, all bins are
            copied, but the bins which are not used are masked.
            Default: False

        Returns
        -------
        df : :class:`~astropy.table.QTable`
            The data contained in the profile.

        Examples
        --------
        >>> sp = ds.sphere("c", (0.1, "unitary"))
        >>> p = sp.profile(
        ...     ("index", "radius"), [("gas", "density"), ("gas", "temperature")]
        ... )
        >>> qt1 = p.to_astropy_table()
        >>> qt2 = p.to_astropy_table(fields=("gas", "density"), only_used=True)
        """
        from astropy.table import QTable

        idxs, masked, fields = self._export_prep(fields, only_used)
        qt = QTable(masked=masked)
        qt[self.x_field[-1]] = self.x[idxs].to_astropy()
        if masked:
            qt[self.x_field[-1]].mask = self.used
        for field in fields:
            qt[field[-1]] = self[field][idxs].to_astropy()
            if masked:
                qt[field[-1]].mask = self.used
        return qt


class Profile1DFromDataset(ProfileNDFromDataset, Profile1D):
    """
    A 1D profile object loaded from a ytdata dataset.
    """

    def __init(self, ds):
        ProfileNDFromDataset.__init__(self, ds)


class Profile2D(ProfileND):
    """An object that represents a 2D profile.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of along the x axis.
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_log : boolean
        Controls whether or not the bins for the x field are evenly
        spaced in linear (False) or log (True) space.
    y_field : string field name
        The field to profile as a function of along the y axis
    y_n : integer
        The number of bins along the y direction.
    y_min : float
        The minimum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_max : float
        The maximum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_log : boolean
        Controls whether or not the bins for the y field are evenly
        spaced in linear (False) or log (True) space.
    weight_field : string field name
        The field to weight the profiled fields by.
    override_bins_x : array
        Array to set as xbins and ignore other parameters if set
    override_bins_y : array
        Array to set as ybins and ignore other parameters if set

    """

    def __init__(
        self,
        data_source,
        x_field,
        x_n,
        x_min,
        x_max,
        x_log,
        y_field,
        y_n,
        y_min,
        y_max,
        y_log,
        weight_field=None,
        override_bins_x=None,
        override_bins_y=None,
    ):
        super().__init__(data_source, weight_field)
        # X
        self.x_field = data_source._determine_fields(x_field)[0]
        self.x_log = x_log
        self.field_info[self.x_field] = self.data_source.ds.field_info[self.x_field]
        x_min, x_max = _sanitize_min_max_units(
            x_min, x_max, self.field_info[self.x_field], self.ds.unit_registry
        )
        self.x_bins = array_like_field(
            data_source, self._get_bins(x_min, x_max, x_n, x_log), self.x_field
        )
        if override_bins_x is not None:
            self.x_bins = array_like_field(data_source, override_bins_x, self.x_field)

        # Y
        self.y_field = data_source._determine_fields(y_field)[0]
        self.y_log = y_log
        self.field_info[self.y_field] = self.data_source.ds.field_info[self.y_field]
        y_min, y_max = _sanitize_min_max_units(
            y_min, y_max, self.field_info[self.y_field], self.ds.unit_registry
        )
        self.y_bins = array_like_field(
            data_source, self._get_bins(y_min, y_max, y_n, y_log), self.y_field
        )
        if override_bins_y is not None:
            self.y_bins = array_like_field(data_source, override_bins_y, self.y_field)

        self.size = (self.x_bins.size - 1, self.y_bins.size - 1)

        self.bin_fields = (self.x_field, self.y_field)
        self.x = 0.5 * (self.x_bins[1:] + self.x_bins[:-1])
        self.y = 0.5 * (self.y_bins[1:] + self.y_bins[:-1])

    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None:
            return
        fdata, wdata, (bf_x, bf_y) = rv
        bf_x.convert_to_units(self.field_info[self.x_field].output_units)
        bin_ind_x = np.digitize(bf_x, self.x_bins) - 1
        bf_y.convert_to_units(self.field_info[self.y_field].output_units)
        bin_ind_y = np.digitize(bf_y, self.y_bins) - 1
        new_bin_profile2d(
            bin_ind_x,
            bin_ind_y,
            wdata,
            fdata,
            storage.weight_values,
            storage.values,
            storage.mvalues,
            storage.qvalues,
            storage.used,
        )
        # We've binned it!

    def set_x_unit(self, new_unit):
        """Sets a new unit for the x field

        Parameters
        ----------
        new_unit : string or Unit object
            The name of the new unit.
        """
        self.x_bins.convert_to_units(new_unit)
        self.x = 0.5 * (self.x_bins[1:] + self.x_bins[:-1])

    def set_y_unit(self, new_unit):
        """Sets a new unit for the y field

        Parameters
        ----------
        new_unit : string or Unit object
            The name of the new unit.
        """
        self.y_bins.convert_to_units(new_unit)
        self.y = 0.5 * (self.y_bins[1:] + self.y_bins[:-1])

    @property
    def bounds(self):
        return ((self.x_bins[0], self.x_bins[-1]), (self.y_bins[0], self.y_bins[-1]))

    def plot(self):
        r"""
        This returns a :class:~yt.visualization.profile_plotter.PhasePlot with
        the fields that have been added to this object.
        """
        from yt.visualization.profile_plotter import PhasePlot

        return PhasePlot.from_profile(self)


class Profile2DFromDataset(ProfileNDFromDataset, Profile2D):
    """
    A 2D profile object loaded from a ytdata dataset.
    """

    def __init(self, ds):
        ProfileNDFromDataset.__init__(self, ds)


class ParticleProfile(Profile2D):
    """An object that represents a *deposited* 2D profile. This is like a
    Profile2D, except that it is intended for particle data. Instead of just
    binning the particles, the added fields will be deposited onto the mesh
    using either the nearest-grid-point or cloud-in-cell interpolation kernels.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of along the x axis.
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    y_field : string field name
        The field to profile as a function of along the y axis
    y_n : integer
        The number of bins along the y direction.
    y_min : float
        The minimum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_max : float
        The maximum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    weight_field : string field name
        The field to use for weighting. Default is None.
    deposition : string, optional
        The interpolation kernel to be used for
        deposition. Valid choices:
        "ngp" : nearest grid point interpolation
        "cic" : cloud-in-cell interpolation

    """

    accumulation = False
    fractional = False

    def __init__(
        self,
        data_source,
        x_field,
        x_n,
        x_min,
        x_max,
        x_log,
        y_field,
        y_n,
        y_min,
        y_max,
        y_log,
        weight_field=None,
        deposition="ngp",
    ):

        x_field = data_source._determine_fields(x_field)[0]
        y_field = data_source._determine_fields(y_field)[0]

        if deposition not in ["ngp", "cic"]:
            raise NotImplementedError(deposition)
        elif (x_log or y_log) and deposition != "ngp":
            mylog.warning(
                "cic deposition is only supported for linear axis "
                "scales, falling back to ngp deposition"
            )
            deposition = "ngp"

        self.deposition = deposition

        # set the log parameters to False (since that doesn't make much sense
        # for deposited data) and also turn off the weight field.
        super().__init__(
            data_source,
            x_field,
            x_n,
            x_min,
            x_max,
            x_log,
            y_field,
            y_n,
            y_min,
            y_max,
            y_log,
            weight_field=weight_field,
        )

    # Either stick the particle field in the nearest bin,
    # or spread it out using the 2D CIC deposition function
    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None:
            return
        fdata, wdata, (bf_x, bf_y) = rv
        # make sure everything has the same units before deposition.
        # the units will be scaled to the correct values later.

        if self.deposition == "ngp":
            func = NGPDeposit_2
        elif self.deposition == "cic":
            func = CICDeposit_2

        for fi, _field in enumerate(fields):
            if self.weight_field is None:
                deposit_vals = fdata[:, fi]
            else:
                deposit_vals = wdata * fdata[:, fi]

            field_mask = np.zeros(self.size, dtype="uint8")

            func(
                bf_x,
                bf_y,
                deposit_vals,
                fdata[:, fi].size,
                storage.values[:, :, fi],
                field_mask,
                self.x_bins,
                self.y_bins,
            )

            locs = field_mask > 0
            storage.used[locs] = True

            if self.weight_field is not None:
                func(
                    bf_x,
                    bf_y,
                    wdata,
                    fdata[:, fi].size,
                    storage.weight_values,
                    field_mask,
                    self.x_bins,
                    self.y_bins,
                )
            else:
                storage.weight_values[locs] = 1.0
            storage.mvalues[locs, fi] = (
                storage.values[locs, fi] / storage.weight_values[locs]
            )
        # We've binned it!


class Profile3D(ProfileND):
    """An object that represents a 2D profile.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of along the x axis.
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_log : boolean
        Controls whether or not the bins for the x field are evenly
        spaced in linear (False) or log (True) space.
    y_field : string field name
        The field to profile as a function of along the y axis
    y_n : integer
        The number of bins along the y direction.
    y_min : float
        The minimum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_max : float
        The maximum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_log : boolean
        Controls whether or not the bins for the y field are evenly
        spaced in linear (False) or log (True) space.
    z_field : string field name
        The field to profile as a function of along the z axis
    z_n : integer
        The number of bins along the z direction.
    z_min : float
        The minimum value of the z profile field. If supplied without units,
        assumed to be in the output units for z_field.
    z_max : float
        The maximum value of thee z profile field. If supplied without units,
        assumed to be in the output units for z_field.
    z_log : boolean
        Controls whether or not the bins for the z field are evenly
        spaced in linear (False) or log (True) space.
    weight_field : string field name
        The field to weight the profiled fields by.
    override_bins_x : array
        Array to set as xbins and ignore other parameters if set
    override_bins_y : array
        Array to set as xbins and ignore other parameters if set
    override_bins_z : array
        Array to set as xbins and ignore other parameters if set

    """

    def __init__(
        self,
        data_source,
        x_field,
        x_n,
        x_min,
        x_max,
        x_log,
        y_field,
        y_n,
        y_min,
        y_max,
        y_log,
        z_field,
        z_n,
        z_min,
        z_max,
        z_log,
        weight_field=None,
        override_bins_x=None,
        override_bins_y=None,
        override_bins_z=None,
    ):
        super().__init__(data_source, weight_field)
        # X
        self.x_field = data_source._determine_fields(x_field)[0]
        self.x_log = x_log
        self.field_info[self.x_field] = self.data_source.ds.field_info[self.x_field]
        x_min, x_max = _sanitize_min_max_units(
            x_min, x_max, self.field_info[self.x_field], self.ds.unit_registry
        )
        self.x_bins = array_like_field(
            data_source, self._get_bins(x_min, x_max, x_n, x_log), self.x_field
        )
        if override_bins_x is not None:
            self.x_bins = array_like_field(data_source, override_bins_x, self.x_field)
        # Y
        self.y_field = data_source._determine_fields(y_field)[0]
        self.y_log = y_log
        self.field_info[self.y_field] = self.data_source.ds.field_info[self.y_field]
        y_min, y_max = _sanitize_min_max_units(
            y_min, y_max, self.field_info[self.y_field], self.ds.unit_registry
        )
        self.y_bins = array_like_field(
            data_source, self._get_bins(y_min, y_max, y_n, y_log), self.y_field
        )
        if override_bins_y is not None:
            self.y_bins = array_like_field(data_source, override_bins_y, self.y_field)
        # Z
        self.z_field = data_source._determine_fields(z_field)[0]
        self.z_log = z_log
        self.field_info[self.z_field] = self.data_source.ds.field_info[self.z_field]
        z_min, z_max = _sanitize_min_max_units(
            z_min, z_max, self.field_info[self.z_field], self.ds.unit_registry
        )
        self.z_bins = array_like_field(
            data_source, self._get_bins(z_min, z_max, z_n, z_log), self.z_field
        )
        if override_bins_z is not None:
            self.z_bins = array_like_field(data_source, override_bins_z, self.z_field)

        self.size = (self.x_bins.size - 1, self.y_bins.size - 1, self.z_bins.size - 1)

        self.bin_fields = (self.x_field, self.y_field, self.z_field)
        self.x = 0.5 * (self.x_bins[1:] + self.x_bins[:-1])
        self.y = 0.5 * (self.y_bins[1:] + self.y_bins[:-1])
        self.z = 0.5 * (self.z_bins[1:] + self.z_bins[:-1])

    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None:
            return
        fdata, wdata, (bf_x, bf_y, bf_z) = rv
        bf_x.convert_to_units(self.field_info[self.x_field].output_units)
        bin_ind_x = np.digitize(bf_x, self.x_bins) - 1
        bf_y.convert_to_units(self.field_info[self.y_field].output_units)
        bin_ind_y = np.digitize(bf_y, self.y_bins) - 1
        bf_z.convert_to_units(self.field_info[self.z_field].output_units)
        bin_ind_z = np.digitize(bf_z, self.z_bins) - 1
        new_bin_profile3d(
            bin_ind_x,
            bin_ind_y,
            bin_ind_z,
            wdata,
            fdata,
            storage.weight_values,
            storage.values,
            storage.mvalues,
            storage.qvalues,
            storage.used,
        )
        # We've binned it!

    @property
    def bounds(self):
        return (
            (self.x_bins[0], self.x_bins[-1]),
            (self.y_bins[0], self.y_bins[-1]),
            (self.z_bins[0], self.z_bins[-1]),
        )

    def set_x_unit(self, new_unit):
        """Sets a new unit for the x field

        Parameters
        ----------
        new_unit : string or Unit object
            The name of the new unit.
        """
        self.x_bins.convert_to_units(new_unit)
        self.x = 0.5 * (self.x_bins[1:] + self.x_bins[:-1])

    def set_y_unit(self, new_unit):
        """Sets a new unit for the y field

        Parameters
        ----------
        new_unit : string or Unit object
            The name of the new unit.
        """
        self.y_bins.convert_to_units(new_unit)
        self.y = 0.5 * (self.y_bins[1:] + self.y_bins[:-1])

    def set_z_unit(self, new_unit):
        """Sets a new unit for the z field

        Parameters
        ----------
        new_unit : string or Unit object
            The name of the new unit.
        """
        self.z_bins.convert_to_units(new_unit)
        self.z = 0.5 * (self.z_bins[1:] + self.z_bins[:-1])


class Profile3DFromDataset(ProfileNDFromDataset, Profile3D):
    """
    A 2D profile object loaded from a ytdata dataset.
    """

    def __init(self, ds):
        ProfileNDFromDataset.__init__(self, ds)


def sanitize_field_tuple_keys(input_dict, data_source):
    if input_dict is not None:
        dummy = {}
        for item in input_dict:
            dummy[data_source._determine_fields(item)[0]] = input_dict[item]
        return dummy
    else:
        return input_dict


def create_profile(
    data_source,
    bin_fields,
    fields,
    n_bins=64,
    extrema=None,
    logs=None,
    units=None,
    weight_field=("gas", "mass"),
    accumulation=False,
    fractional=False,
    deposition="ngp",
    override_bins=None,
):
    r"""
    Create a 1, 2, or 3D profile object.

    The dimensionality of the profile object is chosen by the number of
    fields given in the bin_fields argument.

    Parameters
    ----------
    data_source : YTSelectionContainer Object
        The data object to be profiled.
    bin_fields : list of strings
        List of the binning fields for profiling.
    fields : list of strings
        The fields to be profiled.
    n_bins : int or list of ints
        The number of bins in each dimension.  If None, 64 bins for
        each bin are used for each bin field.
        Default: 64.
    extrema : dict of min, max tuples
        Minimum and maximum values of the bin_fields for the profiles.
        The keys correspond to the field names. Defaults to the extrema
        of the bin_fields of the dataset. If a units dict is provided, extrema
        are understood to be in the units specified in the dictionary.
    logs : dict of boolean values
        Whether or not to log the bin_fields for the profiles.
        The keys correspond to the field names. Defaults to the take_log
        attribute of the field.
    units : dict of strings
        The units of the fields in the profiles, including the bin_fields.
    weight_field : str or tuple field identifier
        The weight field for computing weighted average for the profile
        values.  If None, the profile values are sums of the data in
        each bin. Defaults to ("gas", "mass").
    accumulation : bool or list of bools
        If True, the profile values for a bin n are the cumulative sum of
        all the values from bin 0 to n.  If -True, the sum is reversed so
        that the value for bin n is the cumulative sum from bin N (total bins)
        to n.  If the profile is 2D or 3D, a list of values can be given to
        control the summation in each dimension independently.
        Default: False.
    fractional : bool
        If True the profile values are divided by the sum of all
        the profile data such that the profile represents a probability
        distribution function.
    deposition : strings
        Controls the type of deposition used for ParticlePhasePlots.
        Valid choices are 'ngp' and 'cic'. Default is 'ngp'. This parameter is
        ignored the if the input fields are not of particle type.
    override_bins : dict of bins to profile plot with
        If set, ignores n_bins and extrema settings and uses the
        supplied bins to profile the field. If a units dict is provided,
        bins are understood to be in the units specified in the dictionary.


    Examples
    --------

    Create a 1d profile.  Access bin field from profile.x and field
    data from profile[<field_name>].

    >>> ds = load("DD0046/DD0046")
    >>> ad = ds.all_data()
    >>> profile = create_profile(
    ...     ad, [("gas", "density")], [("gas", "temperature"), ("gas", "velocity_x")]
    ... )
    >>> print(profile.x)
    >>> print(profile["gas", "temperature"])

    """
    bin_fields = data_source._determine_fields(bin_fields)
    fields = list(iter_fields(fields))
    is_pfield = [
        data_source.ds._get_field_info(f).sampling_type == "particle"
        for f in bin_fields + fields
    ]
    wf = None
    if weight_field is not None:
        wf = data_source.ds._get_field_info(weight_field)
        is_pfield.append(wf.sampling_type == "particle")
        wf = wf.name

    if len(bin_fields) > 1 and isinstance(accumulation, bool):
        accumulation = [accumulation for _ in range(len(bin_fields))]

    bin_fields = data_source._determine_fields(bin_fields)
    fields = data_source._determine_fields(fields)
    units = sanitize_field_tuple_keys(units, data_source)
    extrema = sanitize_field_tuple_keys(extrema, data_source)
    logs = sanitize_field_tuple_keys(logs, data_source)
    override_bins = sanitize_field_tuple_keys(override_bins, data_source)

    if any(is_pfield) and not all(is_pfield):
        if hasattr(data_source.ds, "_sph_ptypes"):
            is_local = [
                data_source.ds.field_info[f].sampling_type == "local"
                for f in bin_fields + fields
            ]
            is_local_or_pfield = [pf or lf for (pf, lf) in zip(is_pfield, is_local)]
            if not all(is_local_or_pfield):
                raise YTIllDefinedProfile(
                    bin_fields, data_source._determine_fields(fields), wf, is_pfield
                )
        else:
            raise YTIllDefinedProfile(
                bin_fields, data_source._determine_fields(fields), wf, is_pfield
            )
    if len(bin_fields) == 1:
        cls = Profile1D
    elif len(bin_fields) == 2 and all(is_pfield):
        if deposition == "cic":
            if logs is not None:
                if (bin_fields[0] in logs and logs[bin_fields[0]]) or (
                    bin_fields[1] in logs and logs[bin_fields[1]]
                ):
                    raise RuntimeError(
                        "CIC deposition is only implemented for linear-scaled axes"
                    )
            else:
                logs = {bin_fields[0]: False, bin_fields[1]: False}
            if any(accumulation) or fractional:
                raise RuntimeError(
                    "The accumulation and fractional keyword arguments must be "
                    "False for CIC deposition"
                )
        cls = ParticleProfile
    elif len(bin_fields) == 2:
        cls = Profile2D
    elif len(bin_fields) == 3:
        cls = Profile3D
    else:
        raise NotImplementedError
    if weight_field is not None and cls == ParticleProfile:
        (weight_field,) = data_source._determine_fields([weight_field])
        wf = data_source.ds._get_field_info(weight_field)
        if not wf.sampling_type == "particle":
            weight_field = None
    if not is_sequence(n_bins):
        n_bins = [n_bins] * len(bin_fields)
    if not is_sequence(accumulation):
        accumulation = [accumulation] * len(bin_fields)
    if logs is None:
        logs = {}
    logs_list = []
    for bin_field in bin_fields:
        if bin_field in logs:
            logs_list.append(logs[bin_field])
        else:
            logs_list.append(data_source.ds.field_info[bin_field].take_log)
    logs = logs_list

    # Are the extrema all Nones? Then treat them as though extrema was set as None
    if extrema is None or not any(collapse(extrema.values())):
        ex = [
            data_source.quantities["Extrema"](f, non_zero=l)
            for f, l in zip(bin_fields, logs)
        ]
        # pad extrema by epsilon so cells at bin edges are not excluded
        for i, (mi, ma) in enumerate(ex):
            mi = mi - np.spacing(mi)
            ma = ma + np.spacing(ma)
            ex[i][0], ex[i][1] = mi, ma
    else:
        ex = []
        for bin_field in bin_fields:
            bf_units = data_source.ds.field_info[bin_field].output_units
            try:
                field_ex = list(extrema[bin_field[-1]])
            except KeyError as e:
                try:
                    field_ex = list(extrema[bin_field])
                except KeyError:
                    raise RuntimeError(
                        "Could not find field {} or {} in extrema".format(
                            bin_field[-1], bin_field
                        )
                    ) from e

            if isinstance(field_ex[0], tuple):
                field_ex = [data_source.ds.quan(*f) for f in field_ex]
            if any([exi is None for exi in field_ex]):
                try:
                    ds_extrema = data_source.quantities.extrema(bin_field)
                except AttributeError:
                    # ytdata profile datasets don't have data_source.quantities
                    bf_vals = data_source[bin_field]
                    ds_extrema = data_source.ds.arr([bf_vals.min(), bf_vals.max()])
                for i, exi in enumerate(field_ex):
                    if exi is None:
                        field_ex[i] = ds_extrema[i]
                        # pad extrema by epsilon so cells at bin edges are
                        # not excluded
                        field_ex[i] -= (-1) ** i * np.spacing(field_ex[i])
            if units is not None and bin_field in units:
                for i, exi in enumerate(field_ex):
                    if hasattr(exi, "units"):
                        field_ex[i] = exi.to(units[bin_field])
                    else:
                        field_ex[i] = data_source.ds.quan(exi, units[bin_field])
                fe = data_source.ds.arr(field_ex)
            else:
                if hasattr(field_ex, "units"):
                    fe = field_ex.to(bf_units)
                else:
                    fe = data_source.ds.arr(field_ex, bf_units)
            fe.convert_to_units(bf_units)
            field_ex = [fe[0].v, fe[1].v]
            if is_sequence(field_ex[0]):
                field_ex[0] = data_source.ds.quan(field_ex[0][0], field_ex[0][1])
                field_ex[0] = field_ex[0].in_units(bf_units)
            if is_sequence(field_ex[1]):
                field_ex[1] = data_source.ds.quan(field_ex[1][0], field_ex[1][1])
                field_ex[1] = field_ex[1].in_units(bf_units)
            ex.append(field_ex)

    if override_bins is not None:
        o_bins = []
        for bin_field in bin_fields:
            bf_units = data_source.ds.field_info[bin_field].output_units
            try:
                field_obin = override_bins[bin_field[-1]]
            except KeyError:
                field_obin = override_bins[bin_field]

            if field_obin is None:
                o_bins.append(None)
                continue

            if isinstance(field_obin, tuple):
                field_obin = data_source.ds.arr(*field_obin)

            if units is not None and bin_field in units:
                fe = data_source.ds.arr(field_obin, units[bin_field])
            else:
                if hasattr(field_obin, "units"):
                    fe = field_obin.to(bf_units)
                else:
                    fe = data_source.ds.arr(field_obin, bf_units)
            fe.convert_to_units(bf_units)
            field_obin = fe.d
            o_bins.append(field_obin)

    args = [data_source]
    for f, n, (mi, ma), l in zip(bin_fields, n_bins, ex, logs):
        if mi <= 0 and l:
            raise YTIllDefinedBounds(mi, ma)
        args += [f, n, mi, ma, l]
    kwargs = dict(weight_field=weight_field)
    if cls is ParticleProfile:
        kwargs["deposition"] = deposition
    if override_bins is not None:
        for o_bin, ax in zip(o_bins, ["x", "y", "z"]):
            kwargs[f"override_bins_{ax}"] = o_bin
    obj = cls(*args, **kwargs)
    obj.accumulation = accumulation
    obj.fractional = fractional
    if fields is not None:
        obj.add_fields([field for field in fields])
    for field in fields:
        if fractional:
            obj.field_data[field] /= obj.field_data[field].sum()
        for axis, acc in enumerate(accumulation):
            if not acc:
                continue
            temp = obj.field_data[field]
            temp = np.rollaxis(temp, axis)
            if weight_field is not None:
                temp_weight = obj.weight
                temp_weight = np.rollaxis(temp_weight, axis)
            if acc < 0:
                temp = temp[::-1]
                if weight_field is not None:
                    temp_weight = temp_weight[::-1]
            if weight_field is None:
                temp = temp.cumsum(axis=0)
            else:
                temp = (temp * temp_weight).cumsum(axis=0) / temp_weight.cumsum(axis=0)
            if acc < 0:
                temp = temp[::-1]
                if weight_field is not None:
                    temp_weight = temp_weight[::-1]
            temp = np.rollaxis(temp, axis)
            obj.field_data[field] = temp
            if weight_field is not None:
                temp_weight = np.rollaxis(temp_weight, axis)
                obj.weight = temp_weight
    if units is not None:
        for field, unit in units.items():
            field = data_source._determine_fields(field)[0]
            if field == obj.x_field:
                obj.set_x_unit(unit)
            elif field == getattr(obj, "y_field", None):
                obj.set_y_unit(unit)
            elif field == getattr(obj, "z_field", None):
                obj.set_z_unit(unit)
            else:
                obj.set_field_unit(field, unit)
    return obj



import os.path

import yt
from yt.config import ytcfg
from yt.testing import assert_equal, assert_raises, requires_file

G30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


def setup():
    ytcfg["yt", "serialize"] = True


def teardown():
    ytcfg["yt", "serialize"] = False


@requires_file(G30)
def test_store():
    ds = yt.load(G30)
    store = ds.parameter_filename + ".yt"
    field = "density"
    if os.path.isfile(store):
        os.remove(store)

    proj1 = ds.proj(field, "z")
    sp = ds.sphere(ds.domain_center, (4, "kpc"))
    proj2 = ds.proj(field, "z", data_source=sp)

    proj1_c = ds.proj(field, "z")
    assert_equal(proj1[field], proj1_c[field])

    proj2_c = ds.proj(field, "z", data_source=sp)
    assert_equal(proj2[field], proj2_c[field])

    def fail_for_different_method():
        proj2_c = ds.proj(field, "z", data_source=sp, method="max")
        assert_equal(proj2[field], proj2_c[field])

    # A note here: a unyt.exceptions.UnitOperationError is raised
    # and caught by numpy, which reraises a ValueError
    assert_raises(ValueError, fail_for_different_method)

    def fail_for_different_source():
        sp = ds.sphere(ds.domain_center, (2, "kpc"))
        proj2_c = ds.proj(field, "z", data_source=sp, method="integrate")
        assert_equal(proj2_c[field], proj2[field])

    assert_raises(AssertionError, fail_for_different_source)
